from __future__ import annotations

import asyncio
import ctypes
import os
import threading
import time
from contextlib import asynccontextmanager
from typing import Optional

import sentry_sdk
from fastapi import Body, Depends, FastAPI, Header, HTTPException, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import HTMLResponse
from fastapi.security import HTTPBearer
from fastapi.templating import Jinja2Templates
from github import GithubException
from loguru import logger
from sentry_sdk import set_user

from sweepai.backend.api import app as backend_app
from sweepai.backend.api import fix_issue
from sweepai.backend.api_utils import get_token_header, message_cache
from sweepai.background_jobs import delete_old_repos, scheduler
from sweepai.config.client import SweepConfig
from sweepai.config.server import (
    BLACKLISTED_USERS,
    DISABLED_REPOS,
    ENV,
    GITHUB_LABEL_COLOR,
    GITHUB_LABEL_DESCRIPTION,
    GITHUB_LABEL_NAME,
    PREWARM_REPOS,
    SENTRY_URL,
)
from sweepai.core.entities import Message, PRChangeRequest, Snippet
from sweepai.core.github_utils import get_github_client
from sweepai.handlers.on_comment import on_comment
from sweepai.handlers.on_jira_ticket import handle_jira_ticket
from sweepai.handlers.on_push import on_push
from sweepai.handlers.on_revert_commit_comment import on_revert_commit_comment
from sweepai.handlers.on_ticket import on_ticket
from sweepai.o11y.log_utils import suppress_with_warning
from sweepai.review.auto_suggestions import provide_auto_suggestions
from sweepai.review.review_pr import review_pr
from sweepai.utils.str_utils import BOT_SUFFIX, get_hash
from sweepai.utils.ticket_rendering_utils import PR_REVERT_BUTTON, PR_REVERTED_INDICATOR
from sweepai.web.events import (
    CheckRunCompleted,
    CommentCreatedOrEditedRequest,
    IssueCommentRequest,
    IssueRequest,
    PRLabeledRequest,
    PRRequest,
)
from sweepai.web.global_threads import global_threads
from sweepai.web.health import health_check
from sweepai.web.safe_pqueue import SafePriorityQueue
from sweepai.web.startup import prewarm_search_index
from sweepai.web.sync_clock import sync_time
from sweepai.web.validate_license import validate_license
from sweepai.web.webhook_secret import verify_signature

version = time.strftime("%y.%m.%d.%H")

events = {}
on_ticket_events = {}
review_pr_events = {}
on_push_events = {}

security = HTTPBearer()

templates = Jinja2Templates(directory="sweepai/web")
logger.bind(application="webhook")


@asynccontextmanager
async def lifespan(app: FastAPI):
    if SENTRY_URL:
        sentry_sdk.init(
            dsn=SENTRY_URL,
            traces_sample_rate=1.0,
            profiles_sample_rate=1.0,
            release=version,
        )
    os.makedirs(message_cache, exist_ok=True)
    scheduler.start()
    yield
    with suppress_with_warning(KeyboardInterrupt):
        delete_old_repos()


app = FastAPI(lifespan=lifespan)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(backend_app.router, prefix="/backend")


@app.middleware("http")
async def redirect_chat_backend_routes(request: Request, call_next):
    # For backwards compatibility, redirect /chat/backend/ to /backend/
    if request.url.path.startswith("/chat/backend/"):
        request.scope["path"] = request.url.path.replace("/chat/backend/", "/backend/", 1)
        request.scope["raw_path"] = request.scope["path"].encode()

    response = await call_next(request)
    return response


def run_on_ticket(*args, **kwargs):
    tracking_id = get_hash()
    return asyncio.run(on_ticket(*args, **kwargs, tracking_id=tracking_id))


def run_on_comment(*args, **kwargs):
    tracking_id = get_hash()
    with logger.contextualize(
        **kwargs,
        name="comment_" + kwargs["username"],
        tracking_id=tracking_id,
    ):
        asyncio.run(on_comment(*args, **kwargs, tracking_id=tracking_id))


def run_review_pr(*args, **kwargs):
    tracking_id = get_hash()
    with logger.contextualize(
        **kwargs,
        name="review_" + kwargs["username"],
        tracking_id=tracking_id,
    ):
        review_pr(*args, **kwargs, tracking_id=tracking_id)


def run_on_push(*args, **kwargs):
    tracking_id = get_hash()
    with logger.contextualize(
        **kwargs,
        name="push_" + kwargs["repo_full_name"],
        tracking_id=tracking_id,
    ):
        on_push(*args, **kwargs, tracking_id=tracking_id)


def terminate_thread(thread):
    """Terminate a python threading.Thread."""
    try:
        if not thread.is_alive():
            return

        exc = ctypes.py_object(SystemExit)
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(thread.ident), exc)
        if res == 0:
            raise ValueError("Invalid thread ID")
        elif res != 1:
            # Call with exception set to 0 is needed to cleanup properly.
            ctypes.pythonapi.PyThreadState_SetAsyncExc(thread.ident, 0)
            raise SystemError("PyThreadState_SetAsyncExc failed")
    except Exception as e:
        logger.exception(f"Failed to terminate thread: {e}")


def call_on_ticket(*args, **kwargs):
    global on_ticket_events
    key = f"{kwargs['repo_full_name']}-{kwargs['issue_number']}"  # Full name, issue number as key

    # Use multithreading
    # Check if a previous process exists for the same key, cancel it
    e = on_ticket_events.get(key, None)
    if e:
        logger.info(f"Found previous thread for key {key} and cancelling it")
        terminate_thread(e)

    thread = threading.Thread(target=run_on_ticket, args=args, kwargs=kwargs)
    on_ticket_events[key] = thread
    thread.start()
    global_threads.append(thread)


def call_on_comment(*args, **kwargs):  # TODO: if its a GHA delete all previous GHA and append to the end
    def worker():
        while not events[key].empty():
            task_args, task_kwargs = events[key].get()
            run_on_comment(*task_args, **task_kwargs)

    global events
    repo_full_name = kwargs["repo_full_name"]
    pr_id = kwargs["pr_number"]
    key = f"{repo_full_name}-{pr_id}"  # Full name, comment number as key

    if key not in events:
        events[key] = SafePriorityQueue()

    events[key].put(0, (args, kwargs))

    # If a thread isn't running, start one
    if not any(thread.name == key and thread.is_alive() for thread in threading.enumerate()):
        thread = threading.Thread(target=worker, name=key)
        thread.start()
        global_threads.append(thread)


# add a review by sweep on the pr
def call_review_pr(*args, **kwargs):
    global review_pr_events
    key = f"{kwargs['repository'].full_name}-{kwargs['pr'].number}"  # Full name, issue number as key

    # Use multithreading
    # Check if a previous process exists for the same key, cancel it
    e = review_pr_events.get(key, None)
    if e:
        logger.info(f"Found previous thread for key {key} and cancelling it")
        terminate_thread(e)

    thread = threading.Thread(target=run_review_pr, args=args, kwargs=kwargs)
    review_pr_events[key] = thread
    thread.start()
    global_threads.append(thread)


def call_on_push(*args, **kwargs):
    global on_push_events
    key = kwargs["repo_full_name"]  # Full name as key

    # Use multithreading
    # Check if a previous process exists for the same key, cancel it
    e = on_push_events.get(key, None)
    if e:
        logger.info(f"Found previous thread for key {key} and cancelling it")
        terminate_thread(e)

    thread = threading.Thread(target=run_on_push, args=args, kwargs=kwargs)
    on_push_events[key] = thread
    thread.start()
    global_threads.append(thread)


def call_on_revert_commit_comment(*args, **kwargs):
    tracking_id = get_hash()
    args = (*args, tracking_id)
    thread = threading.Thread(target=on_revert_commit_comment, args=args, kwargs=kwargs)
    thread.start()
    global_threads.append(thread)


@app.get("/health")
def redirect_to_health():
    return health_check()


@app.get("/", response_class=HTMLResponse)
def home(request: Request):
    try:
        validate_license()
        license_expired = False
    except Exception as e:
        logger.warning(e)
        license_expired = True
    return templates.TemplateResponse(
        name="index.html",
        context={
            "version": version,
            "request": request,
            "license_expired": license_expired,
        },
    )


def handle_request(request_dict, event=None):
    """So it can be exported to the listen endpoint."""
    with logger.contextualize(tracking_id="main", env=ENV):
        action = request_dict.get("action")
        try:
            handle_event(request_dict, event)
        except Exception as e:
            logger.exception(str(e))
        logger.info(f"Done handling {event}, {action}")
        return {"success": True}


# @app.post("/")
async def validate_signature(
    request: Request,
    x_hub_signature: Optional[str] = Header(None, alias="X-Hub-Signature-256"),
):
    payload_body = await request.body()
    if not verify_signature(payload_body=payload_body, signature_header=x_hub_signature):
        raise HTTPException(status_code=403, detail="Request signatures didn't match!")


@app.post("/", dependencies=[Depends(validate_signature)])
def webhook(
    request_dict: dict = Body(...),
    x_github_event: Optional[str] = Header(None, alias="X-GitHub-Event"),
):
    """Handle a webhook request from GitHub"""
    with logger.contextualize(tracking_id="main", env=ENV):
        action = request_dict.get("action", None)
        logger.info(f"Received event: {x_github_event}, {action}")
        return handle_request(request_dict, event=x_github_event)


@app.post("/jira")
def jira_webhook(
    request_dict: dict = Body(...),
) -> None:
    def call_jira_ticket(*args, **kwargs):
        thread = threading.Thread(target=handle_jira_ticket, args=args, kwargs=kwargs)
        thread.start()

    call_jira_ticket(event=request_dict)


# Set up cronjob for this
@app.get("/update_sweep_prs_v2")
def update_sweep_prs_v2(repo_full_name: str, installation_id: int):
    # Get a Github client
    _, g = get_github_client(installation_id)

    # Get the repository
    repo = g.get_repo(repo_full_name)
    config = SweepConfig.get_config(repo)

    try:
        branch_ttl = int(config.get("branch_ttl", 7))
    except Exception:
        branch_ttl = 7
    branch_ttl = max(branch_ttl, 1)

    # Get all open pull requests created by Sweep
    pulls = repo.get_pulls(state="open", head="sweep", sort="updated", direction="desc")[:5]

    # For each pull request, attempt to merge the changes from the default branch into the pull request branch
    try:
        for pr in pulls:
            try:
                # make sure it's a sweep ticket
                feature_branch = pr.head.ref
                if not feature_branch.startswith("sweep/") and not feature_branch.startswith("sweep_"):
                    continue
                if "Resolve merge conflicts" in pr.title:
                    continue
                if (
                    pr.mergeable_state != "clean"
                    and (time.time() - pr.created_at.timestamp()) > 60 * 60 * 24
                    and pr.title.startswith("[Sweep Rules]")
                ):
                    pr.edit(state="closed")
                    continue

                repo.merge(
                    feature_branch,
                    pr.base.ref,
                    f"Merge main into {feature_branch}",
                )
            except Exception as e:
                logger.warning(f"Failed to merge changes from default branch into PR #{pr.number}: {e}")
    except Exception:
        logger.warning("Failed to update sweep PRs")


def should_handle_comment(request: CommentCreatedOrEditedRequest | IssueCommentRequest):
    comment = request.comment.body
    return (
        (
            comment.lower().startswith(
                "sweep:"
            )  # we will handle all comments (with or without label) that start with "sweep:" or "@sweep"
            or comment.lower().startswith("@sweep")
        )
        and request.comment.user.type == "User"  # ensure it's a user comment
        and request.comment.user.login not in BLACKLISTED_USERS  # ensure it's not a blacklisted user
        and BOT_SUFFIX not in comment  # we don't handle bot commnents
    )


def should_handle_revert_commit_comment(
    request: IssueCommentRequest | CommentCreatedOrEditedRequest,
):
    previous_body = request.changes.previous_body
    current_body = request.comment.body
    previous_body_has_revert_button = PR_REVERT_BUTTON in previous_body
    current_body_has_reverted_indicator = PR_REVERTED_INDICATOR in current_body
    return previous_body_has_revert_button and current_body_has_reverted_indicator


def handle_event(request_dict, event):
    action = request_dict.get("action")
    username = request_dict.get("sender", {}).get("login")
    if username:
        set_user({"username": username})

    if repo_full_name := request_dict.get("repository", {}).get("full_name"):
        if repo_full_name in DISABLED_REPOS:
            logger.warning(f"Repo {repo_full_name} is disabled")
            return {"success": False, "error_message": "Repo is disabled"}

    match event, action:
        case "pull_request", "opened":
            try:
                pr_request = PRRequest(**request_dict)
                _, g = get_github_client(request_dict["installation"]["id"])
                repo = g.get_repo(request_dict["repository"]["full_name"])
                pr = repo.get_pull(request_dict["pull_request"]["number"])
                # check if review_pr is restricted
                allowed_repos = os.environ.get("PR_REVIEW_REPOS", "")
                allowed_repos_set = set(allowed_repos.split(",")) if allowed_repos else set()
                allowed_usernames = os.environ.get("PR_REVIEW_USERNAMES", "")
                allowed_usernames_set = set(allowed_usernames.split(",")) if allowed_usernames else set()
                # only call review pr if user names are allowed
                # defaults to all users/repos if not set
                if (not allowed_repos or repo.name in allowed_repos_set) and (
                    not allowed_usernames or pr.user.login in allowed_usernames_set
                ):
                    # run pr review
                    call_review_pr(
                        username=pr.user.login,
                        pr=pr,
                        repository=repo,
                        installation_id=pr_request.installation.id,
                    )
                provide_auto_suggestions(pr, tracking_id=get_hash())
            except Exception as e:
                logger.exception(f"Failed to review PR: {e}")
                raise e
        case "pull_request", "labeled":
            try:
                pr_request = PRLabeledRequest(**request_dict)
                # run only if sweep label is added to the pull request
                if GITHUB_LABEL_NAME in [label.name.lower() for label in pr_request.pull_request.labels]:
                    _, g = get_github_client(request_dict["installation"]["id"])
                    repo = g.get_repo(request_dict["repository"]["full_name"])
                    pr = repo.get_pull(request_dict["pull_request"]["number"])

                    # run pr review - no need to check for allowed users/repos if they are adding sweep label
                    call_review_pr(
                        username=pr.user.login,
                        pr=pr,
                        repository=repo,
                        installation_id=pr_request.installation.id,
                    )
                else:
                    logger.info("sweep label not in pull request labels")

            except Exception as e:
                logger.exception(f"Failed to review PR: {e}")
                raise e
        case "issues", "opened":
            request = IssueRequest(**request_dict)
            issue_title_lower = request.issue.title.lower()
            if issue_title_lower.startswith("sweep") or "sweep:" in issue_title_lower or "@sweep" in issue_title_lower:
                _, g = get_github_client(request.installation.id)
                repo = g.get_repo(request.repository.full_name)

                labels = repo.get_labels()
                label_names = [label.name for label in labels]

                if GITHUB_LABEL_NAME not in label_names:
                    try:
                        repo.create_label(
                            name=GITHUB_LABEL_NAME,
                            color=GITHUB_LABEL_COLOR,
                            description=GITHUB_LABEL_DESCRIPTION,
                        )
                    except GithubException as e:
                        if e.status == 422 and any(
                            error.get("code") == "already_exists" for error in e.data.get("errors", [])
                        ):
                            logger.warning(f"Label '{GITHUB_LABEL_NAME}' already exists in the repository")
                        else:
                            raise e
                current_issue = repo.get_issue(number=request.issue.number)
                current_issue.add_to_labels(GITHUB_LABEL_NAME)
        case "issue_comment", "edited":
            request = IssueCommentRequest(**request_dict)
            sweep_labeled_issue = GITHUB_LABEL_NAME in [label.name.lower() for label in request.issue.labels]
            # handle revert comment
            if should_handle_revert_commit_comment(request):
                call_on_revert_commit_comment(request)
            if (
                request.issue is not None
                and sweep_labeled_issue
                and request.comment.user.type == "User"
                and request.comment.user.login not in BLACKLISTED_USERS
                and not request.comment.user.login.startswith("sweep")
                and not (request.issue.pull_request and request.issue.pull_request.url)
            ):
                logger.info("New issue comment edited")
                request.issue.body = request.issue.body or ""
                request.repository.description = request.repository.description or ""

                if not request.comment.body.strip().lower().startswith(GITHUB_LABEL_NAME):
                    logger.info("Comment does not start with 'Sweep', passing")
                    return {
                        "success": True,
                        "reason": "Comment does not start with 'Sweep', passing",
                    }

                call_on_ticket(
                    title=request.issue.title,
                    summary=request.issue.body,
                    issue_number=request.issue.number,
                    username=request.issue.user.login,
                    repo_full_name=request.repository.full_name,
                    installation_id=request.installation.id,
                    is_first_sweep_invocation=False,  # this is an edit
                )
            elif (
                request.issue.pull_request
                and request.comment.user.type == "User"
                and request.comment.user.login not in BLACKLISTED_USERS
            ):
                if should_handle_comment(request):
                    logger.info(f"Handling comment on PR: {request.issue.pull_request}")
                    pr_change_request = PRChangeRequest(
                        params={
                            "repo_full_name": request.repository.full_name,
                            "username": request.comment.user.login,
                            "installation_id": request.installation.id,
                            "pr_number": request.issue.number,
                            "comment_id": request.comment.id,
                        },
                    )
                    call_on_comment(**pr_change_request.params)
        case "issues", "edited":
            request = IssueRequest(**request_dict)
            if (
                GITHUB_LABEL_NAME in [label.name.lower() for label in request.issue.labels]
                and request.sender.type == "User"
                and not request.sender.login.startswith("sweep")
            ):
                logger.info("New issue edited")
                call_on_ticket(
                    title=request.issue.title,
                    summary=request.issue.body,
                    issue_number=request.issue.number,
                    username=request.issue.user.login,
                    repo_full_name=request.repository.full_name,
                    installation_id=request.installation.id,
                    is_first_sweep_invocation=False,  # this is an edit
                )
            else:
                logger.info("Issue edited, but not a sweep issue")
        case "issues", "labeled":
            request = IssueRequest(**request_dict)
            if (
                any(label.name.lower() == GITHUB_LABEL_NAME for label in request.issue.labels)
                and not request.issue.pull_request
            ):
                request.issue.body = request.issue.body or ""
                request.repository.description = request.repository.description or ""
                call_on_ticket(
                    title=request.issue.title,
                    summary=request.issue.body,
                    issue_number=request.issue.number,
                    username=request.issue.user.login,
                    repo_full_name=request.repository.full_name,
                    installation_id=request.installation.id,
                    is_first_sweep_invocation=True,  # issue labeled is the first invocation
                )
        case "issue_comment", "created":
            request = IssueCommentRequest(**request_dict)
            if (
                request.issue is not None
                and GITHUB_LABEL_NAME in [label.name.lower() for label in request.issue.labels]
                and request.comment.user.type == "User"
                and request.comment.user.login not in BLACKLISTED_USERS
                and not (request.issue.pull_request and request.issue.pull_request.url)
                and BOT_SUFFIX not in request.comment.body
            ):
                request.issue.body = request.issue.body or ""
                request.repository.description = request.repository.description or ""

                if not request.comment.body.strip().lower().startswith(GITHUB_LABEL_NAME):
                    logger.info("Comment does not start with 'Sweep', passing")
                    return {
                        "success": True,
                        "reason": "Comment does not start with 'Sweep', passing",
                    }

                call_on_ticket(
                    title=request.issue.title,
                    summary=request.issue.body,
                    issue_number=request.issue.number,
                    username=request.issue.user.login,
                    repo_full_name=request.repository.full_name,
                    installation_id=request.installation.id,
                    is_first_sweep_invocation=False,  # this is a new comment
                )
            elif (
                request.issue.pull_request
                and request.comment.user.type == "User"
                and request.comment.user.login not in BLACKLISTED_USERS
                and BOT_SUFFIX not in request.comment.body
            ):
                if should_handle_comment(request):
                    pr_change_request = PRChangeRequest(
                        params={
                            "repo_full_name": request.repository.full_name,
                            "username": request.comment.user.login,
                            "installation_id": request.installation.id,
                            "pr_number": request.issue.number,
                            "comment_id": request.comment.id,
                        },
                    )
                    call_on_comment(**pr_change_request.params)
        case "pull_request_review_comment", "created":
            request = CommentCreatedOrEditedRequest(**request_dict)
            if should_handle_comment(request):
                pr_change_request = PRChangeRequest(
                    params={
                        "repo_full_name": request.repository.full_name,
                        "username": request.comment.user.login,
                        "installation_id": request.installation.id,
                        "pr_number": request.pull_request.number,
                        "comment_id": request.comment.id,
                    },
                )
                call_on_comment(**pr_change_request.params)
        case "pull_request_review_comment", "edited":
            request = CommentCreatedOrEditedRequest(**request_dict)
            if should_handle_revert_commit_comment(request):
                call_on_revert_commit_comment(request)
            if should_handle_comment(request):
                pr_change_request = PRChangeRequest(
                    params={
                        "repo_full_name": request.repository.full_name,
                        "username": request.comment.user.login,
                        "installation_id": request.installation.id,
                        "pr_number": request.pull_request.number,
                        "comment_id": request.comment.id,
                    },
                )
                call_on_comment(**pr_change_request.params)
        case "check_run", "completed":
            request = CheckRunCompleted(**request_dict)
            pull_request_object = request.check_run.pull_requests[0]
            _, g = get_github_client(request.installation.id)
            repo = g.get_repo(request.repository.full_name)
            pull_request = repo.get_pull(pull_request_object.number)
            provide_auto_suggestions(pull_request, tracking_id=get_hash())
        case "installation_repositories", "added":
            # TODO: do a dummy query
            pass
        case "installation", "created":
            # TODO: do a dummy query
            pass
        case "push", None:
            call_on_push(repo_full_name=repo_full_name)
            # TODO: add this one later when we parse out push events correctly and actually fetch the PR
            # provide_auto_suggestions(pr, tracking_id=get_hash()) # might want to make this one async as well
        case "pull_request", "closed":
            pr_request = PRRequest(**request_dict)
            merged_by = pr_request.pull_request.merged_by.login if pr_request.pull_request.merged_by else None
            if merged_by is not None:
                # TODO: track merged prs here
                # TODO: do a dummy query
                pass
        case "ping", None:
            return {"message": "pong"}
        case _:
            return {"error": "Unsupported type"}


# make this a separate app later, but this is simpler for now


@app.post("/api/fix_issue")
async def fix_issue_endpoint(
    repo_name: str = Body(...),
    query: str = Body(...),
    messages: list[Message] = Body([]),
    snippets: list[Snippet] = Body([]),
    branch: str = Body(None),
    access_token: str = Depends(get_token_header),
):
    try:
        result = await fix_issue(
            repo_full_name=repo_name,
            query=query,
            messages=messages,
            snippets=snippets,
            access_token=access_token,
            branch=branch,
        )
        return result
    except Exception as e:
        logger.error(f"Error in unified_api_endpoint: {str(e)}")
        raise HTTPException(status_code=500, detail=f"An error occurred: {str(e)}")


if __name__ == "__main__":
    sync_time()
    for repo_identifier in PREWARM_REPOS:
        prewarm_search_index(repo_identifier)



# ensure that all additional_messages are 32768 characters at most, if not split them
import os
import re
import subprocess
import textwrap
from dataclasses import dataclass
from itertools import zip_longest
from typing import Callable, get_type_hints

from loguru import logger

from sweepai.config.client import SweepConfig
from sweepai.core.entities import Message, Snippet
from sweepai.core.github_utils import ClonedRepo, MockClonedRepo
from sweepai.core.llm.chat import Thread, continuous_llm_calls
from sweepai.core.llm.convert_openai_anthropic import AnthropicFunctionCall
from sweepai.core.regex_utils import parse_function_call_parameters
from sweepai.dataclasses.llm_state import LLMState
from sweepai.modify.modify_utils import english_join
from sweepai.modify.validate.code_validators import get_parser
from sweepai.search.agent.code_tree import CodeTree
from sweepai.search.agent.ripgrep_utils import cleaned_rg_output
from sweepai.search.query.ticket_utils import prep_snippets
from sweepai.utils.str_utils import find_filenames
from sweepai.utils.streamable_functions import StreamableFunction, streamable

"""
ENTITIES AND ABSTRACTIONS LOGIC FOR TOOL CALLING START
"""


@dataclass
class Parameter:
    description: str
    optional: bool = False


@dataclass
class Tool:
    name: str
    parameters: list[str]
    parameters_explanation: dict[str, str]
    parameters_optional: dict[str, bool]
    function: Callable[..., str] | StreamableFunction[..., str, str]
    description: str = ""

    def get_xml(self, include_function_call_tags: bool = True, include_description: bool = True):
        parameters_xml = "\n".join(
            f"<{parameter}>\n{self.parameters_explanation[parameter]}\n</{parameter}>" for parameter in self.parameters
        )
        function_xml = f"""<{self.name}>
{parameters_xml}
</{self.name}>"""
        if include_function_call_tags:
            function_xml = f"<function_call>\n{function_xml}\n</function_call>"
        if include_description and self.description:
            function_xml = f"{self.name} - {self.description}\n\n{function_xml}"
        return function_xml

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)

    def stream(self, *args, **kwargs):
        return self.function.stream(*args, **kwargs)


def tool(**kwargs):
    """Docstrings are automatically dedented."""

    def decorator(func):
        func_ = func._stream if isinstance(func, StreamableFunction) else func
        type_hints = get_type_hints(func_)
        func_params = func_.__code__.co_varnames[: func_.__code__.co_argcount]
        parameters_explanation = {}
        parameters = []
        parameters_optional = {}
        for parameter in func_params:
            if isinstance(type_hints.get(parameter), Parameter):
                parameters.append(parameter)
                parameters_explanation[parameter] = type_hints[parameter].description
                parameters_optional[parameter] = type_hints[parameter].optional
        docstrings = textwrap.dedent(func_.__doc__ or "").strip()
        return Tool(
            name=func_.__name__ or kwargs.get("name", ""),
            parameters=parameters,
            parameters_explanation=parameters_explanation,
            parameters_optional=parameters_optional,
            function=func,
            description=docstrings,
        )

    return decorator


"""
ENTITIES AND ABSTRACTIONS LOGIC FOR TOOL CALLING END, START OF FUNCTION PARSING LOGIC
"""


def parse_function_calls(response_contents: str, tools: list[Tool]) -> list[dict[str, str]]:
    """
    Parse all tool calls made in the order they appear in the response contents.
    """
    tool_call_parameters = {tool.name: tool.parameters for tool in tools}
    tool_calls = []

    # Create a regex pattern that matches all tool calls
    all_tool_names = "|".join(re.escape(name) for name in tool_call_parameters.keys())
    all_tool_calls_regex = rf"<({all_tool_names})>(?P<function_call>.*?)<\/\1>"

    # Find all tool calls in the order they appear
    all_matches = list(re.finditer(all_tool_calls_regex, response_contents, re.DOTALL))

    # Process matches in the order they were found
    for match in all_matches:
        tool_name = match.group(1)
        tool_call_contents = match.group("function_call")
        parameters = tool_call_parameters[tool_name]
        tool_call = {
            "tool": tool_name,
            "arguments": parse_function_call_parameters(tool_call_contents, parameters),
            "raw_text": match.group(0),  # include the full match
        }
        tool_calls.append(tool_call)

    return tool_calls


def validate_and_parse_function_call(
    function_calls_string: str, thread: Thread, tools: list[Tool]
) -> AnthropicFunctionCall:
    function_calls = parse_function_calls(function_calls_string.strip("\n") + "\n</function_call>", tools)
    if len(function_calls) > 0:
        function_calls[0] = AnthropicFunctionCall(
            function_name=function_calls[0]["tool"],
            function_parameters=function_calls[0]["arguments"],
            raw_text=function_calls[0]["raw_text"],
        )
        if "<function_call>" in function_calls_string:
            thread.messages[-1].content = thread.messages[-1].content.rstrip("\n") + "\n</function_call>"
    return function_calls[0] if len(function_calls) > 0 else None


def validate_and_parse_multi_function_calls(
    function_calls_string: str, thread: Thread, tools: list[Tool]
) -> AnthropicFunctionCall:
    function_calls = parse_function_calls(function_calls_string.strip("\n") + "\n</function_call>", tools)
    for index, function_call in enumerate(function_calls):
        function_calls[index] = AnthropicFunctionCall(
            function_name=function_call["tool"],
            function_parameters=function_call["arguments"],
            raw_text=function_call["raw_text"],
        )
        if function_calls_string.count("<function_call>") < function_calls_string.count("</function_call>"):
            thread.messages[-1].content = thread.messages[-1].content.rstrip("\n") + "\n</function_call>"
    return function_calls


"""
END OF FUNCTION PARSING LOGIC, START OF TOOL CALL HANDLING
"""


@streamable
def handle_function_call(function_call: AnthropicFunctionCall, tools: list[Tool], **kwargs):
    # Use next() instead
    for tool in tools:
        if function_call.function_name == tool.name:
            for param in tool.parameters:
                if param not in function_call.function_parameters and not tool.parameters_optional[param]:
                    return f"ERROR\n\nThe {param} parameter is missing from the function call."
            function_kwargs = {
                **{k: v.strip() if isinstance(v, str) else v for k, v in function_call.function_parameters.items()},
            }
            for param in tool.parameters:
                if param not in function_call.function_parameters:
                    function_kwargs[param] = ""
            function = tool.function._stream if isinstance(tool.function, StreamableFunction) else tool.function
            param_names = function.__code__.co_varnames[: function.__code__.co_argcount]
            for kwarg in kwargs:
                if kwarg in param_names:
                    function_kwargs[kwarg] = kwargs[kwarg]
            if isinstance(tool.function, StreamableFunction):
                for result in tool.function._stream(**function_kwargs):
                    yield result
            else:
                yield tool.function(**function_kwargs)
            break
    else:
        yield "ERROR\n\nInvalid tool name. Must be one of the following: " + ", ".join([tool.name for tool in tools])


@streamable
def handle_multiple_function_calls(function_calls: list[AnthropicFunctionCall], tools: list[Tool], **kwargs) -> dict:
    """
    Handles multiple function calls and returns the results.
    On failure, return immediately without running the rest of the function calls.
    """
    results = ["ERROR" for _ in function_calls]
    did_error = False
    for index, function_call in enumerate(function_calls):
        for result in handle_function_call.stream(function_call, tools, **kwargs):
            results[index] = result
            if result.startswith("ERROR"):
                did_error = True
            yield {
                "status": "success" if not did_error else "error",
                "results": results,
            }

    yield {"status": "success" if not did_error else "error", "results": results}


NO_TOOL_CALL_PROMPT = """FAILURE
Your last function call was incorrectly formatted.

Make sure you provide XML tags for function_call, tool_name and parameters for all function calls. Check the examples section for reference.

Resolve this error by following these steps:
1. In a scratchpad, list the tag name of each XML blocks of your last assistant message.
2. Based on the XML blocks and the contents, determine the last function call you were trying to make.
3. Describe why your last function call was incorrectly formatted.
4. Finally, re-invoke your last function call with the corrected format, with the contents copied over.

Remember to use the following format for your function calls:
<function_calls>
<function_call>
<TOOL_NAME>
<PARAMETER_NAME>PARAMETER_VALUE</PARAMETER_NAME>
...
</TOOL_NAME>
</function_call>
[other function calls]
</function_calls>

The tools available are {tools}."""  # Updated format for explaining how to use funcction calls


def format_multi_function_call_results(
    username: str,
    thread: Thread,
    function_calls_response: dict,
    function_calls: list[AnthropicFunctionCall],
):
    # TODO: use django templates
    formatted_results = ""
    summary_string = ""
    # if no function calls were made, return the user message
    if not function_calls:
        return function_calls_response.get("results", ["No recognized function calls were made!"])[0]
    for index, (function_call, function_call_response) in enumerate(
        zip_longest(function_calls, function_calls_response.get("results", []))
    ):
        # print normally if the function call was successful or if it succeded before the failed call
        if function_call_response is None:
            continue
        if function_calls_response.get("status", "") == "success":
            tool_name = function_call.function_name
            formatted_results += (
                f"<{tool_name}_results>\nYou called the tool {tool_name} with the following parameters:\n\n"
            )
            for param, value in function_call.function_parameters.items():
                if param in function_call.function_parameters:
                    formatted_results += f"<{param}>{value}</{param}>\n"
            # function_call_results = function_calls_response["results"][index]
            formatted_results += (
                f"The tool was used with the following output:\n\n{function_call_response}\n</{tool_name}_results>\n\n"
            )
            # handle the case where the function call failed - update last message

            summary_string += f"\n\n{index}. You called the tool {tool_name} with the following parameters:\n\n"
            for param, value in function_call.function_parameters.items():
                if param in function_call.function_parameters:
                    summary_string += f"{param}: {value}\n"
            summary_string += (
                f"Which was successful, with an output containing {len(function_call_response.splitlines())} lines."
            )
        elif function_calls_response.get("status", "") == "error":
            tool_name = function_call.function_name
            formatted_results += (
                f"<{tool_name}_results>\nYou called the tool {tool_name} with the following parameters:\n\n"
            )
            for param, value in function_call.function_parameters.items():
                if param in function_call.function_parameters:
                    formatted_results += f"{param}: {value}\n"
            # error_message = function_calls_response["message"]
            formatted_results += (
                f"BUT IT FAILED with the following error message:\n{function_call_response}\n</{tool_name}_result>\n\n"
            )
            # update the last message to remove all following function calls after this one
            thread.messages[-1].content = thread.messages[-1].content.split(function_call.raw_text)[0]

            summary_string += f"\n\n{index}. You called the tool {tool_name} with the following parameters:\n\n"
            for param, value in function_call.function_parameters.items():
                if param in function_call.function_parameters:
                    summary_string += f"{param}: {value}\n"
            summary_string += "Which failed."
    formatted_results = formatted_results.strip() + f"\n\n<summary>\n{summary_string.strip()}\n</summary>\n"
    return formatted_results


@streamable
def get_multi_function_calls(
    username: str,
    thread: Thread,
    user_message: str,
    tools: list[Tool],
    llm_kwargs: dict,
    no_tool_call_prompt: str = NO_TOOL_CALL_PROMPT,
    **kwargs,
):
    """
    Get's and handles multiple tool calls per response, returning all function calls and outputs.
    """
    response = ""
    for response in continuous_llm_calls.stream(
        username,
        thread,
        content=user_message,
        stop_sequences=["</function_calls>"],
        **llm_kwargs,
    ):
        # TODO: check for duplicate function calls
        function_calls = validate_and_parse_multi_function_calls(response, thread, tools)
        if "<function_calls>" in response:
            yield response.split("<function_calls>")[0], {}, function_calls
        else:
            yield response.split("<function_call>")[0], {}, function_calls

    function_calls = validate_and_parse_multi_function_calls(response, thread, tools)

    if not function_calls:
        logger.warning("[YELLOW] No function calls were found in the response.")
        user_message = no_tool_call_prompt.format(tools=english_join([tool.name for tool in tools]))
        function_calls_response = {"status": "success", "results": [user_message]}
    else:
        for function_calls_response in handle_multiple_function_calls.stream(function_calls, tools, **kwargs):
            yield response.split("<function_calls>")[0], function_calls_response, function_calls

    if "<function_calls>" in response:
        yield response.split("<function_calls>")[0], function_calls_response, function_calls
    else:
        yield response.split("<function_call>")[0], function_calls_response, function_calls
    return function_calls_response, function_calls


"""
END OF TOOL CALL HANDLING
"""

"""
COMMON PROMPTS
"""
DUPLICATE_QUESTION_MESSAGE = """You've already asked this question: {question}

Please ask a different question. If you can not find the answer in the search results, you need to ask more specific questions or ask questions about tangentially related topics. For example, if you find that a certain functionality is handled in another utilty module, you may need to search for that utility module to find the relevant information."""

SEARCH_RESULT_INSTRUCTIONS = """

First, think step-by-step in a scratchpad to analyze the search results and determine whether the answers provided here are sufficient or if there are additional relevant modules that we may need, such as referenced utility files, docs or tests.

Then, determine if the results are sufficient to answer the user's request:

<request>
{request}
</request>

Option A: Make additional search queries

If the search results are insufficient, you need to ask more specific questions or ask questions about tangentially related topics. For example, if you find that a certain functionality is handled in another utilty module, you may need to search for that utility module to find the relevant information. Keep in mind that you have already asked the following questions, so do not ask them again:

<previously_searched_queries>
{visited_questions}
</previously_searched_queries>

Instead, search for more specific, targetted search queries by analyzing the newly retrieved information. Think step-by-step to construct these better search queries.

Alternatively, if you have found a file that seems relevant but you would like to see the rest of the file, you can use the `access_file` tool to access the contents of the file.

Option B: Submit the task

Otherwise, if you have found all the relevant information to answer the user's request, submit the task using submit_task. If you submit, ensure that the <answer> includes relevant implementations, usages and examples of code wherever possible. Ensure that the <sources> section is MINIMAL and only includes all files you reference in your answer, and is correctly formatted, with each line contains a file path, start line, end line, and justification like in the example in the instructions.

If you have exhausted all search options and still have not found the answer, you can submit an answer with everything you've searched for and found, clarifying that you can not find the answer.

Remember to use the valid function call format for either options."""

RIPGREP_SEARCH_RESULT_INSTRUCTIONS = """

Keep in mind you've already made these previous ripgrep queries, so you shouldn't make them again:
<previously_searched_queries>
{visited_questions}
</previously_searched_queries>"""

DUPLICATE_RIPGREP_MESSAGE = """You've already ran ripgrep on this query: {question}

Please ask a different question. If you can not find the answer in the search results, you need to make a broader ripgrep query."""

"""
COMMON TOOLS
"""


def post_filter_ripgrep_results(
    results: str,
    max_line_length: int = 300,
):
    output = ""
    for line in results.splitlines():
        if len(line) < 300:
            output += line + "\n"
        else:
            output += line[:300] + f"... (omitted {len(line) - 300} characters)\n"
    return output.strip("\n")


"""
TODO: add support for ripgrep cot block
"""


@tool()
def ripgrep(
    justification: Parameter("Justify why this ripgrep query is necessary, and why your search query must exist in the codebase."),  # type: ignore
    query: Parameter("The keyword to search for in the codebase. This will match against all file contents in the repository."),  # type: ignore
    cloned_repo: ClonedRepo,
    llm_state: LLMState,
):
    """
    Search for specific code symbols in the codebase. You should use this instead of `vector_search` if you know the name of the keyword you are searching for, such as for specific user-facing copy or the name of an endpoint. To find all usages of a function or class, you should use `view_entity` instead.

    You may use regex to search through file contents or directly search for a keyword:

    Examples:
    To find where the button containing the user-facing copy "View all" is defined: view all
    To find where the error message "Invalid email" is defined: invalid email

    If you do not see an entity in context, use vector search instead.

    Bad examples:

    To search for where utility functions for push notifications are defined: push_notifications
    Unless you see an invocation of push_notifications, do not use this. It could be called something else like notificationServices.push.

    To search for usages of authentication: authentication
    This is too vague, and the repository might refer to authentication in different ways, such as auth or login. Unless you see it explicitly called like "user.authentication", you should use vector search instead.

    To search for the definition of the class "PushNotificationService": class PushNotificationService
    Use the `view_entity` tool to view all usages of the class.
    """
    if query in llm_state.ripgrep_queries:
        return DUPLICATE_RIPGREP_MESSAGE.format(question=query)
    llm_state.ripgrep_queries.add(query)
    # Future improvement: use globs/regex for both content and filename search
    # Perform content search
    content_results = perform_content_search(query, cloned_repo)
    # Combine and format results
    combined_results = format_results(query, content_results)
    # Add instructions and truncate
    final_output = add_instructions_and_truncate(combined_results, llm_state)

    return f"<ripgrep_response>\n{final_output}\n</ripgrep_response>"


@tool()
def search_for_files(
    justification: Parameter("Explain what files you are trying to find and your rationale."),  # type: ignore
    query: Parameter("The sub path to match all file paths in the repository against. All file paths that contain this sub path will be returned. You can use globs."),  # type: ignore # can be space separated
    cloned_repo: ClonedRepo,
    llm_state: LLMState,
):
    """
    Use this tool to find files in the codebase.
    Search for specific file paths in the repository, this will return all matching file paths.
    Examples:
    For searching for files, use globs:
    To find all files within a directory: <query>path/to/directory/*</query>
    To find all files within all subdirectories: <query>path/to/directory/**</query>
    To find all files that contain a subpath: <query>**/sub/path/**</query>
    To search for multiple files separate the globs/paths with spaces: <query>path/to/directory/* path/to/another/directory/***</query>
    """
    original_query = query
    # Perform filename search
    file_results = perform_filename_search(original_query, cloned_repo)
    formatted_output = ""
    if file_results.strip():
        file_results = truncate_file_results(file_results).strip("\n")
        formatted_output += f"Here are all files that contain '{query}' in the file name:\n\n<file_search_results>\n{file_results}\n</file_search_results>\n"
    else:
        # can be cleaned up later, claude likes to put multiple queries in one search with spaces in between so we enforced it
        for query in original_query.split(" "):
            # see if it tried to search for a file
            FILE_PATH_REGEX = r"(?:^|\b)((?:\.{1,2}/)?(?:[\w.-]+/)*[\w*.-]+\.[\w*-]+)(?=$|\b)"
            file_paths = re.findall(FILE_PATH_REGEX, query)
            if file_paths:
                # automatically search for the directory instead
                directory_subpath = os.path.dirname(file_paths[0])
                # we should not search the directory if there are no / as this has the chance to get all files in the codebase which would destroy the context
                if os.path.sep in directory_subpath:
                    directory_query = f"**/{directory_subpath}/***"
                    new_file_results = perform_filename_search(directory_query, cloned_repo)
                    if new_file_results.strip():
                        new_file_results = truncate_file_results(new_file_results).strip("\n")
                        # TODO: sort based on lexical similarity
                        formatted_output += f"No files were found for '{query}' However, here are all files that contain '{directory_query}' in the file name:\n\n<file_search_results>\n{new_file_results}\n</file_search_results>\n"
                else:
                    formatted_output += f"No files found for '{query}' in the codebase. If you tried to directly search for a file, instead try searching the directory you believe the file is located in. The actual file name may be different from what you assume.\n"
    return formatted_output


def perform_content_search(query: str, cloned_repo: ClonedRepo) -> str:
    if query.startswith('"') and query.endswith('"'):
        query = query[1:-1]
    if query.startswith("'") and query.endswith("'"):
        query = query[1:-1]
    joined_query = [
        "rg",
        "-n",
        "-i",
        "-C=1",
        "--heading",
        "--sort-files",
        f"'{query}'",
        cloned_repo.repo_dir,
    ]
    response = subprocess.run(
        " ".join(joined_query),
        shell=True,
        capture_output=True,
        text=True,
        cwd=cloned_repo.repo_dir,
    )
    if response.returncode == 0:
        results = cleaned_rg_output(
            root_directory=cloned_repo.repo_dir,
            sweep_config=SweepConfig(),
            output=response.stdout,
        )
        return post_filter_ripgrep_results(results)
    return ""


def perform_filename_search(query: str, cloned_repo: ClonedRepo) -> str:
    cleaned_query = query.strip("$")
    file_response = subprocess.run(
        " ".join(
            [
                "rg",
                "--files",
                "--iglob",
                f"'{cleaned_query}'",
            ]
        ),
        shell=True,
        capture_output=True,
        text=True,
        cwd=cloned_repo.repo_dir,
        # executable="rg",
    )
    if file_response.returncode == 0:
        return file_response.stdout
    return ""


def format_results(query: str, content_results: str) -> str:
    formatted_output = ""
    if content_results.strip():
        formatted_output += f"Here are ALL occurrences of '{query}' in the codebase:\n\n```\n{content_results}\n```\n"
    if not formatted_output:
        formatted_output = f"No results found for '{query}' in the codebase."
    return formatted_output


def truncate_file_results(file_results: str, threshold: int = 1000) -> str:
    lines = file_results.splitlines()
    if len(lines) > threshold:
        truncated = "\n".join(lines[:threshold])
        return f"{truncated}\n... (omitted {len(lines) - threshold} lines)"
    return file_results


def add_instructions_and_truncate(results: str, llm_state: LLMState) -> str:
    results += RIPGREP_SEARCH_RESULT_INSTRUCTIONS.format(
        request=llm_state.request,
        visited_questions="\n".join(sorted(list(llm_state.ripgrep_queries))),
    )

    MAX_RIPGREP_LINES = 1000
    result_lines = results.splitlines()
    if len(result_lines) > MAX_RIPGREP_LINES:
        truncated_results = "\n".join(result_lines[:MAX_RIPGREP_LINES])
        return f"{truncated_results}\n... (omitted {len(result_lines) - MAX_RIPGREP_LINES} lines, the massive output implies you need to be more specific in your search query)"
    return results


ACCESS_FILE_PROMPT = """Here are the contents:

<file_contents>
```
{file_contents}
```
</file_contents>

Here are other items in the same directory as {file_path}:
{files_in_directory}{entities_string}"""


@tool()
def access_file(
    justification: Parameter("Justification for why you want to access this file."),  # type: ignore
    file_path: Parameter("The path to the file you want to access. This must not be a file that was previously accessed."),  # type: ignore
    cloned_repo: ClonedRepo,
    llm_state: LLMState,
):
    """
    Displays a skeleton of a file in the codebase and add it to the currently accessed files. Only use this tool if the file has not been accessed before.
    """
    file_path = file_path.lstrip("/")
    viewed_files_key = file_path
    if viewed_files_key in llm_state.viewed_files:
        return f"\nYou have already accessed the preview for file {file_path}.\nTo view specific entities for the file, call view_entity."
    try:
        file_contents = cloned_repo.get_file_contents(file_path)
        # prevent from viewing full file multiple times. if it views the full file, it is not allowed to view it again
        llm_state.viewed_files.add(viewed_files_key)
    except FileNotFoundError as e:
        # this used to be an error, but it happens often enough that it should be a warning (the LLM will try to view a file that doesn't exist)
        logger.warning(f"Could not find file {file_path} access_file:\n{e}")
        similar_files = cloned_repo.get_similar_file_paths(file_path)
        if similar_files:
            return (
                f"ERROR\nThe file {file_path} doesn't exist in this repo, make sure the file path provided is correct.\n\nHere are some similar files in the repo:\n- "
                + "\n- ".join(similar_files)
            )
        else:
            return (
                f"ERROR\nThe file {file_path} doesn't exist in this repo, make sure the file path provided is correct."
            )
    except Exception as e:
        logger.error(f"Error calling access_file:\n{e}")
        return f"ERROR\nAn error occured while trying to access the file {file_path}. {e}"
    num_lines = len(file_contents.splitlines())

    entities_string = ""
    if num_lines > 200:
        language = file_path.split(".")[-1]
        try:
            if language not in ["py", "ts", "tsx", "js", "jsx", "java", "go"]:
                raise Exception(f"Language {language} not supported for file preview")
            get_parser(language)
        except Exception:
            logger.warning(f"Language {language} not supported for file preview, skipping")
        else:
            try:
                tree = CodeTree.from_code(file_contents, language=language)
                preview, entities, unused_entities = tree.get_preview()
                file_contents = preview
                entities_string = ""
                if unused_entities:
                    entities_string += "\n\nWARNING\nThe following entities were not found in the file:\n\n"
                    for entity in sorted(unused_entities):
                        entities_string += f"- {entity}\n"
                if entities:
                    entities_string = "\n\nHere are the collapsed entities in the file:\n\n"
                    for entity, (start_line, end_line) in sorted(entities.items(), key=lambda x: (x[1][0], x[0])):
                        entities_string += f"{entity}: {start_line} - {end_line}\n"
                    entities_string += (
                        "\nDetermine if any of the collapsed entities are relevant. If so, call the `view_entity` tool."
                    )
                    example_entity = list(entities.keys())[0]
                    entities_string += f"\n\nFor example, to view {example_entity} you would call `view_entity` with <entity>{file_path}:{example_entity}</entity>"
            except Exception as e:
                logger.error(f"Error getting preview for file {file_path}: {e}")
                entities_string = ""

    if num_lines > 8000:
        LINES_TO_SHOW = 100
        file_contents = (
            "\n".join(file_contents.splitlines()[:LINES_TO_SHOW]) + f"\n... (omitted {num_lines - LINES_TO_SHOW} lines)"
        )
        return f"Here are the first {LINES_TO_SHOW} lines of the contents:\n<file_contents>\n```\n{file_contents}\n```\n</file_contents>\nThe file {file_path} is too large to display in full. You may use the `access_file` tool to view specific sections of the file."

    try:
        file_extension = os.path.splitext(file_path)[1]
        full_directory_path = os.path.join(cloned_repo.repo_dir, os.path.dirname(file_path))
        items_in_directory = os.listdir(full_directory_path)
        files_in_directory = [f for f in items_in_directory if os.path.isfile(os.path.join(full_directory_path, f))]
        files_in_directory = sorted(files_in_directory, key=lambda x: (x.endswith(file_extension), x.lower()))
        subdirs_in_directory = [f for f in items_in_directory if os.path.isdir(os.path.join(full_directory_path, f))]
        subdirs_in_directory = sorted(subdirs_in_directory)
    except Exception as e:
        logger.error(f"Error getting directory contents for {file_path}: {e}")
        return f"ERROR\nAn error occured while trying to access the file {file_path}. {e}"

    MAX_DIRS = 20
    MAX_FILES = 50
    files_in_directory_string = "\n".join(f"- {f}" for f in files_in_directory[:MAX_FILES]) + "\n"
    if len(files_in_directory) > MAX_FILES:
        files_in_directory_string += f"... (omitted {len(files_in_directory) - MAX_FILES} files)"
    subdirs_in_directory_string = "\n".join(f"- {f}/..." for f in subdirs_in_directory[:MAX_DIRS]) + "\n"
    if len(subdirs_in_directory) > MAX_DIRS:
        subdirs_in_directory_string += f"... (omitted {len(subdirs_in_directory) - MAX_DIRS} directories)"
    return ACCESS_FILE_PROMPT.format(
        file_contents=file_contents,
        file_path=file_path,
        end=num_lines - 1,
        files_in_directory=subdirs_in_directory_string + files_in_directory_string,
        entities_string=entities_string,
    )


"""UTILITY FUNCTIONS"""


def get_function_call_message_base(**kwargs):
    """
    Use this to get the base message for a function call to let the user know what function is being called.
    """
    base_message = {
        "role": "function",
        "content": "",
        "function_call": {
            "function_name": "deciding",
            "function_parameters": "",
            "is_complete": False,
            "thinking": "",
        },
        "key": "function_call_output",
    }

    # Only update existing keys in base_message
    for key, value in kwargs.items():
        if key in base_message:
            if key == "function_call" and isinstance(value, dict):
                # For function_call, only update existing nested keys
                for fc_key, fc_value in value.items():
                    if fc_key in base_message["function_call"]:
                        base_message["function_call"][fc_key] = fc_value
            else:
                base_message[key] = value

    return Message(**base_message)


def get_imports(file_path: str, file_contents: str) -> list[tuple[int, int]]:
    """
    Gets all the line ranges of the import statements in the file.
    """
    import_ranges = []
    try:
        language = file_path.split(".")[-1]
        tree = CodeTree.from_code(file_contents, language=language)
        import_ranges = tree.get_import_line_ranges()
    except Exception as e:
        logger.error(f"Error getting imports: {e}")
        return []
    return import_ranges


def get_preview_and_entities(
    file_path: str, file_contents: str, min_lines: int = 5
) -> tuple[str, dict[str, tuple[int, int]]]:
    """
    Gets all the line ranges of the import statements in the file.
    """
    preview = ""
    entities = {}
    try:
        language = file_path.split(".")[-1]
        tree = CodeTree.from_code(file_contents, language=language)
        preview, entities, unused_entities = tree.get_preview(min_lines=min_lines)
    except Exception as e:
        logger.error(f"Error getting imports: {e}")
        return "", {}, []
    return preview, entities, unused_entities


@streamable
def search_codebase(
    question: str,
    cloned_repo: ClonedRepo,
    k=5,
    *args,
    **kwargs,
):  # -> Generator[tuple, Any, Any]:
    for message, snippets in prep_snippets.stream(cloned_repo, question, skip_analyze_agent=True, k=k, *args, **kwargs):
        yield message, snippets[:k]
    return snippets[:k]


def parse_file_names(function_call_string: str, cloned_repo: ClonedRepo) -> list[Snippet]:
    snippets = []
    try:
        for line in function_call_string.split("\n"):
            if line.startswith("<file_name>"):
                snippet_denotation = line.split("<file_name>")[1].split("</file_name>")[0]
                if ":" in snippet_denotation:
                    file_path, line_numbers = snippet_denotation.rsplit(":")
                    start, end = line_numbers.split("-")
                    file_contents = cloned_repo.get_file_contents(file_path)
                    snippets.append(
                        Snippet(
                            file_path=file_path,
                            start=int(start),
                            end=int(end),
                            content=file_contents,
                            score=1,
                        )
                    )
                else:
                    file_contents = cloned_repo.get_file_contents(file_path)
                    snippets.append(
                        Snippet(
                            file_path=file_path,
                            start=0,
                            end=len(file_contents.splitlines()),
                            content=file_contents,
                            score=1,
                        )
                    )
    except Exception as e:
        logger.error(f"Error parsing file names: {e}")
        return []
    return snippets


def construct_snippet_from_file(file_path: str, cloned_repo: ClonedRepo) -> Snippet:
    file_contents = cloned_repo.get_file_contents(file_path)
    return Snippet(
        content=file_contents,
        start=0,
        end=len(file_contents.splitlines()),
        file_path=file_path,
        score=1.0,
        type_name="source",  # TODO: classify this correctly
    )


def get_all_snippets_from_query(query: str, cloned_repo: ClonedRepo) -> list[Snippet]:
    potential_file_names_in_query: list[str] = find_filenames(query)
    mentioned_files = []
    all_existing_files: set[str] = set(cloned_repo.get_file_list())
    for potential_file_name in potential_file_names_in_query:
        if potential_file_name in all_existing_files:
            mentioned_files.append(construct_snippet_from_file(potential_file_name, cloned_repo))  # noqa: F821
        else:
            # check if it's a suffix of a unique existing file
            all_full_prefix_or_suffix_files = [
                file
                for file in all_existing_files
                if file.endswith(potential_file_name) or potential_file_name.endswith(file)
            ]
            if len(all_full_prefix_or_suffix_files) == 1:
                mentioned_files.append(construct_snippet_from_file(all_full_prefix_or_suffix_files[0], cloned_repo))
    return mentioned_files


if __name__ == "__main__":
    test_tools = [
        Tool("testing", ["abc", "def"], {}, {}, None),
        Tool("abc", ["def", "ghi"], {}, {}, None),
    ]
    #     test_response = """
    #     <function_call>
    # <ask_question_about_codebase>
    # <question>
    # Where is the vector database implementation located in the Sweep AI codebase?
    # </question>
    # </ask_question_about_codebase>
    # </function_call>
    #     """
    #     function_calls = validate_and_parse_multi_function_calls(
    #         test_response,
    #         None,
    #         test_tools
    #     )

    cloned_repo = MockClonedRepo(
        _repo_dir="/tmp/sweep-internal",
        repo_full_name="sweepai/sweep-internal",
    )

    llm_state = LLMState(
        request="",
        visited_snippets=set(),
        visited_questions=set(),
        visited_ripgrep=set(),
        ripgrep_queries=set(),
        viewed_files=set(),
    )
    # results = ripgrep("app.tsx", cloned_repo, llm_state)
    # results2 = search_for_files("testing", "sweepai/agents/*", cloned_repo, llm_state)
    file_path = "sweepai/utils/github_utils.py"
    entities = ""
    results = access_file("testing", file_path, entities, cloned_repo, llm_state)
    print(results)



"""
Take a PR and provide an AI generated review of the PR.
"""

import copy
import multiprocessing
import os
import re

import git
from github.PullRequest import PullRequest
from loguru import logger
from numpy import argmax, array
from tqdm import tqdm

from sweepai.config.client import SweepConfig
from sweepai.core.entities import Message, UnsuitableFileException
from sweepai.core.github_utils import (
    ClonedRepo,
    MockClonedRepo,
    get_review_threads,
    update_file,
)
from sweepai.core.llm.chat import LATEST_CLAUDE_MODEL, Thread
from sweepai.dataclasses.codereview import (
    CodeReview,
    CodeReviewByGroup,
    CodeReviewIssue,
    FunctionDef,
    GroupedFilesForReview,
    Patch,
    PRChange,
    PRReviewComment,
    PRReviewCommentThread,
)
from sweepai.o11y.event_logger import posthog
from sweepai.review.review_annotations import get_diff_annotations
from sweepai.review.review_prompts import (
    comment_format,
    comment_thread_format,
    patch_format,
    patch_format_without_annotations,
    patches_for_file_format,
    pr_review_change_group_format,
    source_code_with_patches_format,
    system_prompt,
    system_prompt_edge_case_answer,
    system_prompt_edge_case_question_creation,
    system_prompt_identify_new_functions,
    system_prompt_identify_repeats,
    system_prompt_pr_summary,
    system_prompt_review,
    system_prompt_sort_issues,
    system_prompt_special_rules,
    user_prompt,
    user_prompt_edge_case_question_answer_format,
    user_prompt_edge_case_question_creation_format,
    user_prompt_identify_new_functions,
    user_prompt_identify_repeats,
    user_prompt_issue_output_format,
    user_prompt_pr_summary,
    user_prompt_review_analysis_format,
    user_prompt_review_decisions,
    user_prompt_review_questions,
    user_prompt_review_special_rules,
    user_prompt_sort_issues,
    user_prompt_special_rules_format,
)
from sweepai.search.query.ticket_utils import get_top_k_snippets
from sweepai.search.query.vector_db import cosine_similarity, embed_text_array
from sweepai.utils.cache import file_cache
from sweepai.utils.file_utils import read_file_with_fallback_encodings, safe_decode
from sweepai.utils.str_utils import (
    add_line_numbers,
    extract_object_fields_from_string,
    extract_objects_from_string,
    object_to_xml,
    objects_to_xml,
    remove_lines_from_text,
)
from sweepai.utils.ticket_rendering_utils import parse_issues_from_code_review
from sweepai.web.webhook_secret import hash_sha256

# approximately 120k tokens - this is an underestimate which is intentional, even if we can use 128k context we dont want to use all of it
MAX_CHAR_BUDGET = 120000 * 3.5


def get_pr_diffs(pr: PullRequest):
    repo = pr.base.repo

    base_sha = pr.base.sha
    head_sha = pr.head.sha

    comparison = repo.compare(base_sha, head_sha)
    return comparison.files

def validate_diff(file_name: str, diff: str):
    MAX_DIFF_LENGTH = 1000
    if not diff:
        return "diff is empty", False
    if not file_name:
        return "file name is empty", False
    if len(diff.split("\n")) > MAX_DIFF_LENGTH:
        return f"diff is over {MAX_DIFF_LENGTH}", False
    return "", True

@file_cache()
def get_pr_changes(
    pr: PullRequest, cloned_repo: ClonedRepo
) -> tuple[dict[str, PRChange], list[str], list[str]]:
    repo = pr.base.repo

    sweep_config: SweepConfig = SweepConfig()
    base_sha = pr.base.sha
    head_sha = pr.head.sha

    comparison = repo.compare(base_sha, head_sha)
    file_diffs = comparison.files

    pr_diffs = {}
    dropped_files = []  # files that were dropped due them being commonly ignored
    unsuitable_files: list[tuple[str, Exception]] = (
        []
    )  # files dropped for other reasons such as being way to large or not encodable, error objects included
    for file in tqdm(file_diffs, desc="Annotating diffs"):
        file_name = file.filename
        diff = file.patch
        # Ensure diff is a string
        if not isinstance(diff, str):
            diff = str(diff)

        # we can later migrate this to use a cloned repo and fetch off of two hashes
        reason, is_valid_diff = validate_diff(file_name, diff)
        if not is_valid_diff:
            logger.info(f"Skipping invalid diff for file {file_name} because {reason}")
            continue
        previous_filename = file.previous_filename or file.filename

        # drop excluded files: for example package-lock.json files
        if sweep_config.is_file_excluded(file_name):
            dropped_files.append(file_name)
            continue

        errored = False
        e = None
        if file.status == "added":
            old_code = ""
        else:
            try:
                old_code = safe_decode(repo=repo, path=previous_filename, ref=base_sha)
                if old_code is None:
                    raise UnsuitableFileException("Could not decode file")
            except Exception as e_:
                e = e_
                errored = True
                unsuitable_files.append((file_name, e))
        if file.status == "removed":
            new_code = ""
        else:
            try:
                new_code = safe_decode(repo=repo, path=file.filename, ref=head_sha)
                if new_code is None:
                    raise UnsuitableFileException("Could not decode file")
            except Exception as e_:
                e = e_
                errored = True
                unsuitable_files.append((file_name, e))

        # drop unsuitable files
        if new_code:
            suitable, reason = sweep_config.is_file_suitable(new_code)
            if not suitable:
                errored = True
                e = UnsuitableFileException(reason)
                unsuitable_files.append((file_name, e))
            else:
                # final pass using github linguist to check if the file is bad or not
                auto_generated, reason = sweep_config.is_file_bad(file_name, cloned_repo.repo_dir)
                if auto_generated:
                    errored = True
                    e = UnsuitableFileException(reason)
                    unsuitable_files.append((file_name, e))

        if errored:
            posthog.capture(
                "get_pr_changes",
                "get_pr_changes error",
                properties={
                    "error": str(e),
                    "file_name": file_name,
                    "pr": str(pr),
                    "repo": str(repo),
                },
            )
            continue

        status = file.status
        pr_change = PRChange(
            file_name=file_name,
            diff=diff,
            old_code=old_code,
            new_code=new_code,
            status=status,
            patches=split_diff_into_patches(diff, file_name),
        )
        diff_annotations = get_diff_annotations(
            source_code=pr_change.new_code,
            diffs=[patch.changes for patch in pr_change.patches],
            file_name=pr_change.file_name,
        )
        pr_change.annotations = diff_annotations
        pr_diffs[file_name] = pr_change
    return pr_diffs, dropped_files, unsuitable_files


def split_diff_into_patches(diff: str, file_name: str) -> list[Patch]:
    patches = []
    hunks = re.findall(r"@@ -\d+,\d+ \+\d+,\d+ @@.*?(?=\n@@ -|\Z)", diff, re.DOTALL)
    for hunk in hunks:
        line_numbers = re.findall(r"-(\d+),(\d+) \+(\d+),(\d+)", hunk)
        if line_numbers:
            old_start, old_count, new_start, new_count = map(int, line_numbers[0])
            changes = hunk[hunk.index("@@") :].strip()
            patch = Patch(
                file_name=file_name,
                old_start=old_start,
                old_count=old_count,
                new_start=new_start,
                new_count=new_count,
                changes=changes,
            )
            patches.append(patch)
    return patches


# format only the patches for the PRChange
def format_patches_for_pr_change(pr_change: PRChange, include_patch_annotations: bool = True):
    patches = ""
    for idx, patch in enumerate(pr_change.patches):
        if include_patch_annotations:
            patches += patch_format.format(
                file_name=pr_change.file_name,
                index=idx + 1,
                diff=patch.changes,
                annotation=pr_change.annotations[idx],
            )
        else:
            patches += patch_format_without_annotations.format(
                file_name=pr_change.file_name,
                index=idx + 1,
                diff=patch.changes,
            )
        if idx < len(pr_change.patches) - 1:
            patches += "\n"
    formatted_patches = patches_for_file_format.format(file_name=pr_change.file_name, file_patches=patches)
    return formatted_patches


# prunes a file based on the patches for that file, removes long sections in between
def smart_prune_file_based_on_patches(file_contents: str, patches: list[Patch], context_lines: int = 10):
    if not file_contents:
        return file_contents
    lines = file_contents.splitlines(keepends=True)
    num_of_lines = len(lines)
    patch_ranges = []
    # if we dont have patches we naively truncate the file to around 100 lines
    if len(patches) == 0:
        if len(lines) > 100:
            return "".join(lines[:100]) + "...\n"
        else:
            return file_contents

    # sort patches based on new_start
    sorted_patches = sorted(patches, key=lambda patch: patch.new_start)
    for patch in sorted_patches:
        start = max(0, patch.new_start - context_lines - 1)
        end = min(num_of_lines - 1, patch.new_start + patch.new_count + context_lines - 1)
        if len(patch_ranges) == 0:
            patch_ranges.append((start, end))
        else:
            previous_range = patch_ranges[-1]
            # combine if overlap
            if previous_range[1] >= start:
                patch_ranges[-1] = (previous_range[0], end)
            else:
                patch_ranges.append((start, end))
    # now replace any sections not within the patch ranges with ...
    new_lines = []
    for i, range in enumerate(patch_ranges):
        range_lines = lines[range[0] : range[1] + 1]
        # with list slicing we can safely extend past the actual length of the array
        if range[0] != 0:
            range_lines = ["...\n"] + range_lines
        # check if we need trailing ...
        if i == len(patch_ranges) - 1 and range[1] < len(lines) - 1:
            range_lines = range_lines + ["...\n"]
        new_lines.extend(range_lines)
    new_file_contents = "".join(new_lines)
    return new_file_contents


def format_pr_changes(
    pr_changes: dict[str, PRChange],
    include_annotations: bool = True,
    truncate: bool = False,
) -> str:
    formatted_changes = ""
    all_formatted_patches = ""
    all_formatted_source_code = ""
    for file_name, pr_change in pr_changes.items():
        # format patches
        formatted_patches_for_file = format_patches_for_pr_change(
            pr_change, include_patch_annotations=include_annotations
        )
        all_formatted_patches += formatted_patches_for_file
        # format source code
        if truncate:
            formatted_source_code = source_code_with_patches_format.format(
                file_name=file_name,
                file_contents=smart_prune_file_based_on_patches(
                    add_line_numbers(pr_change.new_code, start=1), pr_change.patches
                ),
            )
        else:
            formatted_source_code = source_code_with_patches_format.format(
                file_name=file_name,
                file_contents=add_line_numbers(pr_change.new_code, start=1),
            )
        all_formatted_source_code += formatted_source_code
    file_names = ", ".join(pr_changes.keys())
    formatted_changes = pr_review_change_group_format.format(
        file_names=file_names,
        all_patches=all_formatted_patches,
        all_source_code=all_formatted_source_code,
    )
    return formatted_changes, all_formatted_patches, all_formatted_source_code


# render a group of pr changes for review
def render_pr_changes(
    pr_changes: dict[str, PRChange],
    truncate: bool = False,
    include_annotations: bool = True,
) -> str:
    formatted_changes, formatted_patches, formatted_source_code = format_pr_changes(
        pr_changes, truncate=truncate, include_annotations=include_annotations
    )
    if len(formatted_changes) >= MAX_CHAR_BUDGET:
        truncate = True
        # go again with pruned code
        formatted_changes, formatted_patches, formatted_source_code = format_pr_changes(
            pr_changes, truncate=truncate, include_annotations=include_annotations
        )
    if len(formatted_changes) >= MAX_CHAR_BUDGET:
        truncate = True
        include_annotations = False
        # go again without annotations and prune code
        formatted_changes, formatted_patches, formatted_source_code = format_pr_changes(
            pr_changes, truncate=truncate, include_annotations=include_annotations
        )
    if len(formatted_changes) >= MAX_CHAR_BUDGET:
        # simply too many changes to handle, split them up
        formatted_changes, formatted_patches, formatted_source_code = "", "", ""
    return formatted_changes, formatted_patches, formatted_source_code


# render all changes for all groups, handle cases where grouped files are too large
def format_all_pr_changes_by_groups(
    grouped_files: dict[str, list[str]],
    pr_changes: dict[str, PRChange],
    truncate: bool = False,
    include_annotations: bool = True,
):
    formatted_pr_changes_by_groups: dict[str, GroupedFilesForReview] = {}
    for _, files_to_review in grouped_files.items():
        # get all relevant PRChange objects
        changes_to_review = {file_name: pr_changes[file_name] for file_name in files_to_review}
        # build change overview
        formatted_changes, formatted_patches, formatted_source_code = render_pr_changes(
            changes_to_review,
            truncate=truncate,
            include_annotations=include_annotations,
        )
        if formatted_changes:
            group = GroupedFilesForReview(
                file_names=files_to_review,
                rendered_changes=formatted_changes,
                rendered_patches=formatted_patches,
                rendered_source_code=formatted_source_code,
            )
            formatted_pr_changes_by_groups[group.group_name] = group
        else:  # too large break up
            for file_name in files_to_review:
                changes_to_review = {file_name: pr_changes[file_name]}
                formatted_changes, formatted_patches, formatted_source_code = render_pr_changes(changes_to_review)
                group = GroupedFilesForReview(
                    file_names=[file_name],
                    rendered_changes=formatted_changes,
                    rendered_patches=formatted_patches,
                    rendered_source_code=formatted_source_code,
                )
                formatted_pr_changes_by_groups[group.group_name] = group
    return formatted_pr_changes_by_groups


CLAUDE_MODEL = LATEST_CLAUDE_MODEL
CLAUDE_MODEL_OPUS = "claude-3-opus-20240229"


class PRReviewBot(Thread):
    # check if there are special .md rules to see if there are additional rules to check for
    # assumes by default the file name is gpt.md
    def get_special_rules(
        self,
        cloned_repo: ClonedRepo,
        file_names: list[str],
        special_rule_file: str = "SWEEP.md",
    ):
        # filename: special rules
        special_rules: dict[str, list[str]] = {}
        # a set of already added SWEEP.md rules so that we don't add them multiple times
        added_rules = set([])
        for file_name in file_names:
            # ensure file exists and this is not a GPT.md file
            full_path = os.path.join(cloned_repo.repo_dir, file_name)
            base_file = os.path.basename(file_name)
            if not os.path.exists(full_path) or base_file == special_rule_file:
                logger.error(f"Failure fetching special rules file for {file_name} as it does not exist.")
                return special_rules

            subdirectories = file_name.split(os.path.sep)
            directories_to_check = []
            # Loop from the end to the beginning, excluding the file itself
            for i in range(len(subdirectories) - 1, 0, -1):
                directories_to_check.append(os.path.join(cloned_repo.repo_dir, os.path.join(*subdirectories[:i])))
            # append the base directory
            directories_to_check.append(cloned_repo.repo_dir)
            # now check all directories for the special rule file, we add on to special rules if we find it
            for directory in directories_to_check:
                special_rule_path = os.path.join(directory, special_rule_file)
                if os.path.exists(special_rule_path):
                    if special_rule_path not in added_rules:
                        if file_name not in special_rules:
                            special_rules[file_name] = []
                        special_rules[file_name].append(read_file_with_fallback_encodings(special_rule_path))
                        added_rules.add(special_rule_path)

        return special_rules

    # get a comprehensive pr summary
    def get_pr_summary(
        self,
        formatted_patches: str,
        pull_request_info: str = "",
    ):
        self.messages = [
            Message(
                role="system",
                content=system_prompt_pr_summary,
            )
        ]
        formatted_user_prompt = user_prompt_pr_summary.format(
            all_patches=formatted_patches, pull_request_info=pull_request_info
        )
        pr_summary_response = self.chat(
            content=formatted_user_prompt,
            temperature=0.1,
            model=CLAUDE_MODEL,
        )
        pr_summary = ""
        pr_summary_pattern = r"<pr_summary>(?P<pr_summary>.*?)</pr_summary>"
        pr_summary_match = re.search(pr_summary_pattern, pr_summary_response, re.DOTALL)
        if pr_summary_match:
            pr_summary = pr_summary_match.group("pr_summary")
        return pr_summary

    # fetch all potential issues for each file based on the diffs of that file
    def review_code_changes_by_file(
        self,
        pr_changes: dict[str, PRChange],
        formatted_pr_changes_by_group: dict[str, GroupedFilesForReview],
        formatted_pr_changes_by_file: dict[str, GroupedFilesForReview],
        cloned_repo: ClonedRepo,
        pull_request_info: str,
        seed: int | None = None,
    ):
        code_reviews_by_group = {}
        # loop through all groups
        for group_name, grouped_files in formatted_pr_changes_by_group.items():
            file_names = grouped_files.file_names
            rendered_changes = grouped_files.rendered_changes
            # get all relevant PRChange objects
            # build prompt
            self.messages = [
                Message(
                    role="system",
                    content=system_prompt,
                )
            ]
            formatted_user_prompt = user_prompt.format(
                diff=rendered_changes,
                pull_request_info=pull_request_info,
            )
            formatted_user_prompt += user_prompt_issue_output_format
            if len(formatted_user_prompt) > MAX_CHAR_BUDGET:
                # if we exceed the budget we need to prune the file
                posthog.capture(
                    "review_code_changes_by_file",
                    "review_code_changes_by_file budget exceeded",
                    properties={
                        "file_names": file_names,
                        "body": formatted_user_prompt,
                    },
                )
            code_review_response = self.chat(
                content=formatted_user_prompt,
                temperature=0,
                model=CLAUDE_MODEL,
                seed=seed,
            )
            # make a seperate call for the special rules
            # check if there are special rules we need to follow for this file by seeing if the files "SWEEP.md" exists
            special_rules = self.get_special_rules(cloned_repo, file_names)
            if special_rules:
                # may need to multi process this
                code_review_response = self.check_special_rules(
                    formatted_pr_changes_by_file,
                    code_review_response,
                    special_rules,
                    seed=seed,
                )
            # add a final check for edge cases
            code_review_response = self.check_edge_cases(rendered_changes, code_review_response, seed=seed)

            diff_summary = ""
            diff_summary_pattern = r"<diff_summary>(.*?)</diff_summary>"
            diff_summary_matches = re.findall(diff_summary_pattern, code_review_response, re.DOTALL)
            if diff_summary_matches:
                # join all of them into a single string
                diff_summary = "\n".join([match.strip() for match in diff_summary_matches])
            issues = ""
            issues_pattern = r"<issues>(.*?)</issues>"
            issues_matches = re.findall(issues_pattern, code_review_response, re.DOTALL)
            if issues_matches:
                issues = "\n".join([match.strip() for match in issues_matches])
            potential_issues = parse_issues_from_code_review(issues)
            code_reviews_by_group[group_name] = CodeReviewByGroup(
                file_names=file_names,
                diff_summary=diff_summary,
                issues=potential_issues,
                potential_issues=[],
            )
        return code_reviews_by_group

    def check_special_rules(
        self,
        formatted_pr_changes_by_file: dict[str, GroupedFilesForReview],
        code_review_response: str,
        special_rules: dict[str, list[str]],
        seed: int | None = None,
    ):
        for file_name, special_rule_set in special_rules.items():
            for special_rule in special_rule_set:
                rendered_file_changes = formatted_pr_changes_by_file[file_name].rendered_changes
                formatted_user_prompt_special_rules = user_prompt_special_rules_format.format(
                    diff=rendered_file_changes, special_rules=special_rule
                )
                formatted_user_prompt_special_rules += user_prompt_issue_output_format
                self.messages = [
                    Message(
                        role="system",
                        content=system_prompt_special_rules,
                    )
                ]
                special_rules_response = self.chat(
                    content=formatted_user_prompt_special_rules,
                    temperature=0.1,
                    model=CLAUDE_MODEL_OPUS,
                    seed=seed,
                )
                code_review_response += special_rules_response
        return code_review_response

    def check_edge_cases(
        self,
        rendered_changes: str,
        code_review_response: str,
        seed: int | None = None,
    ):
        self.messages = [
            Message(
                role="system",
                content=system_prompt_edge_case_question_creation,
            )
        ]

        formatted_user_prompt_edge_case_question_creation = user_prompt_edge_case_question_creation_format.format(
            diff=rendered_changes
        )

        edge_case_questions_response = self.chat(
            content=formatted_user_prompt_edge_case_question_creation,
            temperature=0.1,
            model=CLAUDE_MODEL,
            seed=seed,
        )

        created_questions_dict, failed, _ = extract_object_fields_from_string(
            edge_case_questions_response, ["created_questions"]
        )
        if not failed and "# No relevant changes detected" not in edge_case_questions_response:
            created_questions = created_questions_dict["created_questions"]
            # now get llm to answer the questions it created
            self.messages = [
                Message(
                    role="system",
                    content=system_prompt_edge_case_answer,
                )
            ]
            formatted_user_prompt_edge_case_answer = user_prompt_edge_case_question_answer_format.format(
                diff=rendered_changes, questions=created_questions
            )
            formatted_user_prompt_edge_case_answer += user_prompt_issue_output_format
            edge_case_answer_response = self.chat(
                content=formatted_user_prompt_edge_case_answer,
                temperature=0.1,
                model=CLAUDE_MODEL,
                seed=seed,
            )
            code_review_response += edge_case_answer_response
        return code_review_response

    # review the generated issues more critically for each file to see if they are actually important or not
    def review_code_issues_by_file(
        self,
        pr_changes: dict[str, PRChange],
        formatted_pr_changes_by_group: dict[str, GroupedFilesForReview],
        code_reviews_by_group: dict[str, CodeReviewByGroup],
        cloned_repo: ClonedRepo,
        pull_request_info: str,
        formatted_comment_threads: dict[str, str],
        pull_request_summary: str,
        seed: int | None = None,
    ):
        # go file by file
        for group_name, code_review_by_group in code_reviews_by_group.items():
            file_names = code_review_by_group.file_names
            comment_threads = []
            # build comment threads based on the file names
            for file_name in file_names:
                if file_name in formatted_comment_threads:
                    comment_threads.append(formatted_comment_threads[file_name])
            comment_threads_string = "\n".join(comment_threads)
            self.messages = [
                Message(
                    role="system",
                    content=system_prompt_review,
                )
            ]
            # if no issues were identified continue to next file
            if not code_review_by_group.issues:
                continue
            # convert our CodeReviewIssue list to an xml string
            potential_issues_string = objects_to_xml(
                code_review_by_group.issues,
                "issue",
                outer_field_name="potential_issues",
            )
            # now prepend all other pr changes to the current pr change
            all_other_pr_changes = "\n\n".join(
                [
                    file_group.rendered_patches
                    for group, file_group in formatted_pr_changes_by_group.items()
                    if group != group_name
                ]
            )
            # create user prompt
            formatted_user_prompt = user_prompt_review_questions.format(
                file_names=code_review_by_group.get_all_file_names(),
                potential_issues=potential_issues_string,
                pull_request_info=pull_request_info,
                pr_changes=f"{all_other_pr_changes}\n\n{formatted_pr_changes_by_group[group_name].rendered_changes}",
                comment_threads=comment_threads_string,
                pull_request_summary=pull_request_summary,
            )
            special_rules = self.get_special_rules(cloned_repo, file_names)
            if special_rules:
                #  this can be improved by doing it sepeartely
                special_rules_set = set([])
                for special_rule_set in special_rules.values():
                    for special_rule in special_rule_set:
                        special_rules_set.add(special_rule)
                formatted_user_prompt += user_prompt_review_special_rules.format(
                    special_rules="\n".join(special_rules_set)
                )
            formatted_user_prompt += user_prompt_review_analysis_format
            formatted_user_prompt += user_prompt_review_decisions
            # get response
            code_review_response = self.chat(
                content=formatted_user_prompt,
                temperature=0.1,
                model=CLAUDE_MODEL,
                seed=seed,
            )

            severe_issues_pattern = r"<severe_issues>(.*?)</severe_issues>"
            issues_matches = re.findall(severe_issues_pattern, code_review_response, re.DOTALL)
            if issues_matches:
                issues = "\n".join([match.strip() for match in issues_matches])
                potential_issues = parse_issues_from_code_review(issues)
            else:
                potential_issues = []

            # update the issues
            code_reviews_by_group[group_name].issues = potential_issues
        return code_reviews_by_group

    # given a list of changes identify newly created functions
    def identify_functions_in_patches(
        self,
        pr_changes: dict[str, PRChange],
    ):
        newly_created_functions: dict[str, list[FunctionDef]] = {}
        files_to_patches: dict[str, str] = {}
        files_to_pr_change: dict[str, PRChange] = {}
        # format all patches for all files
        for file_name, pr_change in pr_changes.items():
            patches = format_patches_for_pr_change(pr_change)
            files_to_patches[file_name] = patches
            files_to_pr_change[file_name] = pr_change
        # go file by file
        for file_name, patches in files_to_patches.items():
            if "SWEEP.md" in file_name:  # jank but temporary
                continue
            pr_change = files_to_pr_change[file_name]
            self.messages = [
                Message(
                    role="system",
                    content=system_prompt_identify_new_functions,
                )
            ]
            numbered_code_file = add_line_numbers(files_to_pr_change[file_name].new_code, start=1)
            # only need code related to patches
            numbered_code_file = smart_prune_file_based_on_patches(
                numbered_code_file,
                files_to_pr_change[file_name].patches,
                context_lines=5,
            )

            formatted_user_prompt = user_prompt_identify_new_functions.format(
                file_name=file_name,
                patches=patches,
                numbered_code_file=numbered_code_file,
            )
            new_functions_response = self.chat(
                content=formatted_user_prompt,
                temperature=0,
                model=CLAUDE_MODEL,
            )
            # extract function defs from string
            function_def_params = ["function_code", "start_line", "end_line"]
            newly_created_functions_regex = r"<newly_created_functions>(?P<content>.*?)<\/newly_created_functions>"
            newly_created_functions_match = re.search(newly_created_functions_regex, new_functions_response, re.DOTALL)
            if newly_created_functions_match:
                extracted_functions, _ = extract_objects_from_string(
                    newly_created_functions_match.group("content"),
                    "function",
                    function_def_params,
                )
                patches = pr_change.patches
                for extracted_function in extracted_functions:
                    # do some basic double checking, make sure the start and end lines make sense
                    # the start and end lines should fall within the start and end of one patch, if they dont, then it is clearly wrong
                    try:
                        start = int(extracted_function.get("start_line", -1))
                        end = int(extracted_function.get("end_line", -1))
                    except ValueError as e:  # invalid start and end lines
                        logger.error(
                            "Non fatal error in identify_functions_in_patches attempting to extract start and end lines."
                        )
                        posthog.capture(
                            "identify_repeated_functions",
                            "identify_repeated_functions error line_numbers",
                            properties={
                                "error": str(e),
                                "extracted_function": str(extracted_function),
                            },
                        )
                        start = -1
                        end = -1
                    if start != -1 and end != -1:
                        valid_function = False
                        for patch in patches:
                            if start >= patch.new_start and end <= (patch.new_start + patch.new_count):
                                valid_function = True
                                break
                        if valid_function:
                            if file_name not in newly_created_functions:
                                newly_created_functions[file_name] = []
                            newly_created_functions[file_name].append(
                                FunctionDef(**{**extracted_function, "file_name": file_name})
                            )
                        else:
                            logger.warning(
                                f"Extracted function was dropped due to incorrect start and end lines!\nFunction:\n{extracted_function}"
                            )
            else:
                newly_created_functions[file_name] = []
        return newly_created_functions

    # identifies any repeated utility function definitons and raises them as codereviewissue
    def identify_repeated_functions(
        self,
        cloned_repo: ClonedRepo,
        newly_created_functions_dict: dict[str, list[FunctionDef]],
        pull_request_info: str,
        formatted_comment_threads: dict[str, str],
    ) -> dict[str, list[CodeReviewIssue]]:
        repeated_functions_code_issues: dict[str, list[CodeReviewIssue]] = {}
        for file_name, newly_created_functions in newly_created_functions_dict.items():
            if "SWEEP.md" in file_name:  # jank but temporary
                continue
            comment_threads_string = formatted_comment_threads.get(file_name, "")
            # keep copy of edited files to revert later
            modified_files_dict: dict[str, dict[str, str]] = {}
            modified_files_dict[file_name] = {"original": cloned_repo.get_file_contents(file_name)}
            repeated_functions_code_issues[file_name] = []
            # do a similarity search over the chunked code base to see if the function name matches anything
            for function in newly_created_functions:
                # remove the function definition from the file to prevent biased results
                modified_files_dict[file_name]["modified"] = remove_lines_from_text(
                    modified_files_dict[file_name]["original"],
                    start=int(function.start_line),
                    end=int(function.end_line),
                )
                # now update the cloned repo file in both repo_dir
                try:
                    update_file(
                        cloned_repo.repo_dir,
                        file_name,
                        modified_files_dict[file_name]["modified"],
                    )
                except Exception as e:
                    logger.error(f"Failure updating file {cloned_repo.repo_dir}{function.file_name}: {e}")
                    posthog.capture(
                        "identify_repeated_functions",
                        "identify_repeated_functions error updating",
                        properties={
                            "error": str(e),
                            "cloned_repo.repo_dir": cloned_repo.repo_dir,
                            "file_name": function.file_name,
                        },
                    )
                    raise e
                file_hash = hash_sha256(modified_files_dict[file_name]["modified"])
                # get the top five snippets and then pass those into sweep to ask if there are any repeated function definitions
                _, ranked_snippets, _, _ = get_top_k_snippets(
                    cloned_repo,
                    function.function_code,
                    k=3,
                    do_not_use_file_cache=True,
                    use_repo_dir=True,
                    seed=file_hash,
                )
                formatted_code_snippets = "\n\n".join(
                    [
                        f"<code_snippet file_name='{snippet.file_path}' snippet_index='{idx}'>\n{snippet.get_snippet()}\n</code_snippet>"
                        for idx, snippet in enumerate(ranked_snippets)
                    ]
                )
                formatted_user_prompt = user_prompt_identify_repeats.format(
                    function=f"<new_function>\n{function.function_code}\n</new_function>",
                    formatted_code_snippets=formatted_code_snippets,
                    pull_request_info=pull_request_info,
                    comment_threads=comment_threads_string,
                )
                self.messages = [
                    Message(
                        role="system",
                        content=system_prompt_identify_repeats,
                    )
                ]
                repeated_functions_response = self.chat(
                    content=formatted_user_prompt,
                    temperature=0.1,
                    model=CLAUDE_MODEL,
                )
                repeated_function_params = ["answer", "justification"]
                repeated_function, _, failed_param = extract_object_fields_from_string(
                    repeated_functions_response, repeated_function_params
                )
                # if extraction fails
                if failed_param == "answer":
                    logger.error("Failure in extract_object_fields_from_string")
                    posthog.capture(
                        "extract_object_fields_from_string",
                        "extract_object_fields_from_string failed",
                        properties={
                            "text": repeated_functions_response,
                            "params": str(repeated_function_params),
                        },
                    )
                else:
                    # case insensitive match
                    answer_true = r"true"
                    if bool(re.search(answer_true, repeated_function["answer"], re.IGNORECASE)):
                        justification = repeated_function.get("justification", "")
                        new_code_issue = CodeReviewIssue(
                            file_name=file_name,
                            issue_description=f"Sweep has identified a redundant function: {justification}",
                            line_number=function.start_line,
                        )
                        repeated_functions_code_issues[function.file_name].append(new_code_issue)

                # now revert the cloned repo file - if this fails this can cause big issues
                try:
                    update_file(
                        cloned_repo.repo_dir,
                        file_name,
                        modified_files_dict[file_name]["original"],
                    )
                except Exception as e:
                    logger.error(f"Failure updating file {cloned_repo.repo_dir}{function.file_name}: {e}")
                    posthog.capture(
                        "identify_repeated_functions",
                        "identify_repeated_functions error reverting",
                        properties={
                            "error": str(e),
                            "cloned_repo.repo_dir": cloned_repo.repo_dir,
                            "file_name": function.file_name,
                        },
                    )
                    raise e
        return repeated_functions_code_issues

    # sorts issues by severity, potential issues are not sorted
    # returns modified code_review_by_file and a list of all issues sorted
    def sort_code_issues_by_severity(
        self,
        code_review_by_file: dict[str, CodeReview],
    ) -> tuple[dict[str, CodeReview], list[CodeReviewIssue]]:
        all_issues_formatted = "<all_issues>\n"
        index_to_issues: dict[int, CodeReviewIssue] = {}
        issue_index = 1
        # build all_issues string
        for file_name, code_review in code_review_by_file.items():
            all_issues = code_review.issues
            for issue in all_issues:
                all_issues_formatted += f"\n<issue index='{issue_index}'>\n{issue.issue_description}\n</issue>\n"
                index_to_issues[issue_index] = issue
                issue_index += 1
        all_issues_formatted += "\n</all_issues>"

        # if there was only one minus issue there is no need to sort anything
        if issue_index <= 2:
            return code_review_by_file, []

        self.messages = [
            Message(
                role="system",
                content=system_prompt_sort_issues,
            )
        ]
        formatted_user_prompt = user_prompt_sort_issues.format(all_issues=all_issues_formatted)
        sorted_issues_response = self.chat(
            content=formatted_user_prompt,
            temperature=0.1,
            model=CLAUDE_MODEL,
        )

        sorted_issue_indices, _, _ = extract_object_fields_from_string(
            sorted_issues_response, ["sorted_issue_indices_by_severity"]
        )
        # parse output to reorder the issues
        sorted_indices_string = sorted_issue_indices["sorted_issue_indices_by_severity"].split(",")
        sorted_indices = []
        for index in sorted_indices_string:
            try:
                index = int(index.strip())
                _issue = index_to_issues[index]
                sorted_indices.append(index)
            except Exception as e:
                logger.error(f"Failure parsing indices in sort_code_issues_by_severity: {e}")
        # sort the issues within the CodeReview object but also create a global list of all issues
        all_issues_sorted: list[CodeReviewIssue] = []
        for index in sorted_indices:
            all_issues_sorted.append(index_to_issues[index])
        # now rebuild the issue arrays for each of the code reviews
        for file_name, code_review in code_review_by_file.items():
            code_review.issues = []
            for issue in all_issues_sorted:
                if issue.file_name == file_name:
                    code_review.issues.append(issue)
        return code_review_by_file, all_issues_sorted


# get the best issue to return based on group vote
def get_group_voted_best_issue_index(
    username: str,
    file_name: str,
    label: str,
    files_to_labels_indexes: dict[str, dict[str, list[int]]],
    files_to_embeddings: dict[str, any],
    index_length: int,
):
    similarity_scores = [0 for _ in range(index_length)]
    for index_i in range(index_length):
        for index_j in range(index_length):
            if index_i != index_j:
                embedding_i = files_to_embeddings[file_name][index_i].reshape(1, 512)
                embedding_j = files_to_embeddings[file_name][index_j].reshape(1, 512)
                similarity_scores[index_i] += cosine_similarity(embedding_i, embedding_j)[0][0]
    max_index = files_to_labels_indexes[file_name][label][argmax(similarity_scores)]
    return max_index


# function that gets the code review for every file in the pr
def get_code_reviews_for_file(
    pr_changes: dict[str, PRChange],
    formatted_pr_changes_by_group: dict[str, GroupedFilesForReview],
    formatted_pr_changes_by_file: dict[str, GroupedFilesForReview],  # single file
    cloned_repo: ClonedRepo,
    pull_request_info: str,
    formatted_comment_threads: dict[str, str],
    pull_request_summary: str,
    seed: int | None = None,
):
    review_bot = PRReviewBot()
    code_review_by_group = review_bot.review_code_changes_by_file(
        pr_changes,
        formatted_pr_changes_by_group,
        formatted_pr_changes_by_file,
        cloned_repo,
        pull_request_info,
        seed=seed,
    )
    code_review_by_group = review_bot.review_code_issues_by_file(
        pr_changes,
        formatted_pr_changes_by_group,
        code_review_by_group,
        cloned_repo,
        pull_request_info,
        formatted_comment_threads,
        pull_request_summary,
        seed=seed,
    )
    return code_review_by_group


# run 5 seperate instances of review_pr and then group the resulting issues and only take the issues that appear the majority of the time (> 3)
def group_vote_review_pr(
    username: str,
    pr_changes: dict[str, PRChange],
    formatted_pr_changes_by_group: dict[str, GroupedFilesForReview],
    formatted_pr_changes_by_file: dict[str, GroupedFilesForReview],  # this one is all single files
    cloned_repo: ClonedRepo,
    pull_request_info: str,
    formatted_comment_threads: dict[str, str],
    pull_request_summary: str,
    multiprocess: bool = True,
) -> dict[str, CodeReview]:
    majority_code_review_by_file = {}
    code_reviews_by_group: list[dict[str, CodeReviewByGroup]] = []
    GROUP_SIZE = 5
    if multiprocess:
        pool = multiprocessing.Pool(processes=5)
        # we have to create copies of the ClonedRepo or else the repo_dir will get cleaned up when the processes terminate
        cloned_repos = [
            MockClonedRepo(
                _repo_dir=cloned_repo.repo_dir,
                repo_full_name=cloned_repo.repo_full_name,
                installation_id=cloned_repo.installation_id,
                branch=cloned_repo.branch,
                token=cloned_repo.token,
                repo=cloned_repo.github_repo,
                git_repo=git.Repo(cloned_repo.repo_dir),
            )
            for _ in range(GROUP_SIZE)
        ]
        results = [
            pool.apply_async(
                get_code_reviews_for_file,
                args=(
                    pr_changes,
                    formatted_pr_changes_by_group,
                    formatted_pr_changes_by_file,
                    cloned_repos[i],
                    pull_request_info,
                    formatted_comment_threads,
                    pull_request_summary,
                    i,
                ),
            )
            for i in range(GROUP_SIZE)
        ]
        pool.close()
        pool.join()
        for result in results:
            try:
                code_review = result.get()
                code_reviews_by_group.append(code_review)
            except Exception as e:
                logger.error(f"Error fetching result: {e}")
                posthog.capture(
                    username,
                    "get_code_reviews_for_file error multiprocess",
                    properties={"error": str(e)},
                )
    else:
        for i in range(GROUP_SIZE):
            review = get_code_reviews_for_file(
                pr_changes,
                formatted_pr_changes_by_group,
                formatted_pr_changes_by_file,
                cloned_repo,
                pull_request_info,
                formatted_comment_threads,
                pull_request_summary,
                seed=i,
            )
            code_reviews_by_group.append(review)
    # embed each issue and then cluster them
    # extract code issues for each file and prepare them for embedding
    code_reviews_ready_for_embedding = []
    for code_review_by_group in code_reviews_by_group:
        prepped_code_review: dict[str, list[str]] = {}
        for group_name, code_review in code_review_by_group.items():
            # using object_to_xml may not be the most optimal as it adds extra xml tags
            prepped_code_review[group_name] = [object_to_xml(code_issue, "issue") for code_issue in code_review.issues]
        code_reviews_ready_for_embedding.append(prepped_code_review)

    # embed all extracted texts
    code_reviews_embeddings = []
    for prepped_code_review in code_reviews_ready_for_embedding:
        embedded_code_review: dict[str, list] = {}
        for group_name, code_issues in prepped_code_review.items():
            embedded_code_review[group_name] = embed_text_array(code_issues)
        code_reviews_embeddings.append(embedded_code_review)
    # dbscan - density based spatial clustering of app with noise
    # format: {file_name: [label1, label2, ...]}
    groups_to_labels = {}
    # corresponding issues for each file
    # format: {file_name: [issue1, issue2, ...]}
    groups_to_issues = {}
    # corresponding embeddings for each file
    # format: {file_name: [embedding1, embedding2, ...]}
    groups_to_embeddings = {}

    # for each group combine all the embeddings together while determining the max amount of clusters
    for group_name in formatted_pr_changes_by_group:
        all_embeddings = []
        all_issues = []
        for i in range(len(code_reviews_by_group)):
            embeddings = code_reviews_embeddings[i][group_name]
            code_review = code_reviews_by_group[i][group_name]
            if embeddings:
                embeddings = embeddings[0]
                for embedding in embeddings:
                    all_embeddings.append(embedding.flatten())
                all_issues.extend(code_review.issues)
        groups_to_issues[group_name] = all_issues
        all_flattened_embeddings = array(all_embeddings)
        groups_to_embeddings[group_name] = all_flattened_embeddings
        # note DBSCAN expects a shape with less than or equal to 2 dimensions
        try:
            if all_flattened_embeddings.size:
                from sklearn.cluster import DBSCAN

                db = DBSCAN(eps=0.375, min_samples=2).fit(all_flattened_embeddings)
                groups_to_labels[group_name] = db.labels_
            else:
                groups_to_labels[group_name] = []
        except ValueError as e:
            logger.error(f"Error with dbscan {e}")
    LABEL_THRESHOLD = 4
    # get the labels that have a count greater than the threshold
    # format: {file_name: {label: [index, ...]}}
    groups_to_labels_indexes = {}
    for group_name, labels in groups_to_labels.items():
        index_dict: dict[str, list[int]] = {}
        for i, v in enumerate(labels):
            key = str(v)
            if key not in index_dict:
                index_dict[key] = []
            index_dict[key].append(i)
        groups_to_labels_indexes[group_name] = index_dict

    # create the final code_reviews_by_file
    for group_name, labels_dict in groups_to_labels_indexes.items():
        # pick first one as diff summary doesnt really matter
        final_code_review: CodeReview = copy.deepcopy(code_reviews_by_group[0][group_name])
        final_code_review.issues = []
        final_code_review.potential_issues = []
        final_issues = []
        potential_issues = []
        for label, indexes in labels_dict.items():
            index_length = len(indexes)
            # -1 is considered as noise
            if index_length >= LABEL_THRESHOLD and label != "-1":
                max_index = get_group_voted_best_issue_index(
                    username,
                    group_name,
                    label,
                    groups_to_labels_indexes,
                    groups_to_embeddings,
                    index_length,
                )
                # add to final issues, first issue - TODO use similarity score of all issues against each other
                final_issues.append(groups_to_issues[group_name][max_index])
            # get potential issues which are one below the label_threshold
            if index_length == LABEL_THRESHOLD - 1 and label != "-1":
                max_index = get_group_voted_best_issue_index(
                    username,
                    group_name,
                    label,
                    groups_to_labels_indexes,
                    groups_to_embeddings,
                    index_length,
                )
                potential_issues.append(groups_to_issues[group_name][max_index])
        final_code_review.issues = final_issues
        final_code_review.potential_issues = potential_issues
        majority_code_review_by_file[group_name] = copy.deepcopy(final_code_review)
    return majority_code_review_by_file


def review_pr_detailed_checks(
    username: str,
    cloned_repo: ClonedRepo,
    pr_changes: dict[str, PRChange],
    code_review_by_file: dict[str, CodeReview],
    pull_request_info: str,
    formatted_comment_threads: dict[str, str],
) -> dict[str, CodeReview]:
    review_bot = PRReviewBot()
    # get a list of newly created functions
    newly_created_functions_dict: dict[str, list[FunctionDef]] = review_bot.identify_functions_in_patches(pr_changes)
    new_code_issues: dict[str, list[CodeReviewIssue]] = review_bot.identify_repeated_functions(
        cloned_repo,
        newly_created_functions_dict,
        pull_request_info,
        formatted_comment_threads,
    )
    # now append these code issues to the existing ones
    for file_name, new_code_issues in new_code_issues.items():
        if new_code_issues:
            code_review_by_file[file_name].issues.extend(new_code_issues)

    return code_review_by_file


# get the summary for a pr given all the changes
def get_pr_summary_from_patches(
    pr_changes: dict[str, PRChange],
    pull_request_info: str = "",
):
    review_bot = PRReviewBot()
    formatted_pr_patches = ""
    for file_name, pr_change in pr_changes.items():
        patches = format_patches_for_pr_change(pr_change)
        formatted_pr_patches += f'\n\n<patches file_name="{file_name}">\n{patches}\n</patches>\n\n'
    pr_summary = review_bot.get_pr_summary(formatted_pr_patches, pull_request_info=pull_request_info)
    return pr_summary


def sort_code_issues_by_severity(
    username: str,
    code_review_by_file: dict[str, CodeReview],
) -> dict[str, CodeReview]:
    review_bot = PRReviewBot()
    MAX_ISSUE_AMOUNT = 10
    # sort all the issues by severity
    code_review_by_file, all_issues_sorted = review_bot.sort_code_issues_by_severity(code_review_by_file)
    all_issues_sorted = all_issues_sorted[:MAX_ISSUE_AMOUNT]

    return code_review_by_file, all_issues_sorted


def format_pr_info(pr: PullRequest) -> str:
    info = ""
    try:
        title = pr.title
        if title:
            info += f"<pr_title>\n{title}\n<\pr_title>\n\n"
    except Exception as e:
        logger.warning(f"Couldn't fetch title for pr: {pr}\nError: {e}")
    try:
        description = pr.body
        if description:
            info += f"<pr_description>\n{description}\n</pr_description>\n\n"
    except Exception as e:
        logger.warning(f"Couldn't fetch body for pr: {pr}\nError: {e}")
    return info


# cluster patches based on similarity and review based off of that
def cluster_patches(pr_changes: dict[str, PRChange]):
    all_patches_by_file: dict[str, list[Patch]] = {}
    all_patch_strings_by_file: dict[str, str] = {}
    for file_name, pr_change in pr_changes.items():
        all_patches_by_file[file_name] = pr_change.patches
        all_patch_strings_by_file[file_name] = objects_to_xml(
            pr_change.patches,
            "patch",
            "patches",
            exclude_fields=["old_start", "new_start", "old_count", "new_count"],
        )
    # get the order of the files
    file_order = [file_name for file_name in all_patch_strings_by_file.keys()]
    files_to_embed = [patches for patches in all_patch_strings_by_file.values()]
    try:
        embedded_patches = embed_text_array(files_to_embed)[0]
        from sklearn.cluster import DBSCAN

        db = DBSCAN(eps=0.6, min_samples=2).fit(embedded_patches)
        labels = db.labels_
    except Exception:
        # if we cant cluster for some reason just set each file in its own group
        labels = [i for i in range(len(file_order))]

    # group key is the label, value is the list of file names to review together
    groups_to_review_files_in: dict[str, list[str]] = {}
    # split files into their groups -> -1 means create a seperate group -1x for that group
    noise_groups = 0
    for index, group in enumerate(labels):
        if group != -1:
            group_key = str(group)
        else:
            group_key = str(group) + str(noise_groups)
            noise_groups += 1
        if group_key not in groups_to_review_files_in:
            groups_to_review_files_in[group_key] = []
        groups_to_review_files_in[group_key].append(file_order[index])

    return groups_to_review_files_in


def decompose_code_review_by_group(code_reviews_by_group: dict[str, CodeReviewByGroup]):
    code_review_by_file: dict[str, CodeReview] = {}
    for group_name, code_review_by_group in code_reviews_by_group.items():
        file_names = code_review_by_group.file_names
        for file_name in file_names:
            issues = [issue for issue in code_review_by_group.issues if issue.file_name == file_name]
            potential_issues = [
                issue for issue in code_review_by_group.potential_issues if issue.file_name == file_name
            ]
            code_review = CodeReview(
                file_name=file_name,
                diff_summary=code_review_by_group.diff_summary,
                issues=issues,
                potential_issues=potential_issues,
            )
            code_review_by_file[file_name] = code_review
    return code_review_by_file


def get_all_comments_for_review(
    repo_full_name: str, pr: PullRequest, installation_id: int
) -> dict[str, list[PRReviewCommentThread]]:
    review_threads = get_review_threads(repo_full_name, pr.number, installation_id)
    # build a comments to review thread dictionary
    # {comment_id: {thread_id, isResolved}}
    file_name_to_comments: dict[str, list[PRReviewCommentThread]] = {}
    for review_thread in review_threads:
        thread_id = review_thread["id"]
        is_resolved = review_thread["isResolved"]
        is_outdated = review_thread["isOutdated"]
        line = review_thread["line"]
        path = review_thread["path"]
        thread_comments = review_thread["comments"]["nodes"]
        pr_thread_review_comments: list[PRReviewComment] = []
        for thread_comment in thread_comments:
            pr_review_comment: PRReviewComment = PRReviewComment(
                thread_id=thread_id,
                file_name=path,
                line_number=line,
                body=thread_comment["body"],
                author=thread_comment["author"]["login"],
                is_outdated=is_outdated,
                is_resolved=is_resolved,
            )
            pr_thread_review_comments.append(pr_review_comment)

        pr_comment_thread = PRReviewCommentThread(
            thread_id=thread_id,
            file_name=path,
            line_number=line,
            is_resolved=is_resolved,
            is_outdated=is_outdated,
            comments=pr_thread_review_comments,
        )
        if path not in file_name_to_comments:
            file_name_to_comments[path] = []
        file_name_to_comments[path].append(pr_comment_thread)

    return file_name_to_comments


def render_comment_thread(comment_thread: PRReviewCommentThread) -> str:
    rendered_comment_thread = ""
    # if it is outdated there is no need to render it
    if comment_thread.is_outdated:
        return rendered_comment_thread
    all_comments = comment_thread.comments
    all_comments_formatted = []
    for comment in all_comments:
        formatted_comment = comment_format.format(
            file_name=comment.file_name,
            line_number=comment.line_number,
            comment_body=comment.body,
        )
        all_comments_formatted.append(formatted_comment)
    rendered_comment_thread = comment_thread_format.format(
        file_name=comment_thread.file_name,
        line_number=comment_thread.line_number,
        is_resolved=comment_thread.is_resolved,
        comments="\n".join(all_comments_formatted),
    )
    # if comment_thread.is_resolved:
    #     rendered_comment_thread = comment_thread_is_resolved_format.format(
    #         comment_thread=rendered_comment_thread
    #     )
    return rendered_comment_thread


# format all comment threads for use in pr review
def format_comment_threads(comment_threads: dict[str, list[PRReviewCommentThread]]) -> dict[str, str]:
    formatted_comment_threads: dict[str, str] = {}

    # format all comment threads per file
    for file_name, comment_threads in comment_threads.items():
        rendered_comment_threads = (
            f"The file {file_name} had the following comment threads addressing the issues you previously raised:"
        )
        for comment_thread in comment_threads:
            rendered_comment_threads += f"{render_comment_thread(comment_thread)}"
        if file_name not in formatted_comment_threads:
            formatted_comment_threads[file_name] = ""
        formatted_comment_threads[file_name] += rendered_comment_threads
    return formatted_comment_threads



from datetime import datetime
from github.PullRequest import PullRequest
from loguru import logger
import pytz

from sweepai.ci.fetch_ci import get_ci_failure_logs, get_ci_failures
from sweepai.config.server import FIX_CI_ALLOWLIST
from sweepai.core.github_utils import github_integration
from sweepai.core.llm.chat import LATEST_CLAUDE_MODEL, Thread, continuous_llm_calls
from sweepai.o11y.log_utils import log_to_file
from sweepai.utils.diff import remove_whitespace_changes
from sweepai.utils.format_utils import Prompt
from sweepai.utils.str_utils import extract_all_xml_tags, wildcard_match
from sweepai.utils.ticket_rendering_utils import create_message_marker, create_or_edit_sweeps_github_comment, find_sweeps_github_comment

system_message = "You are a helpful assistant that reviews pull requests and suggests unit tests that should be added to a pull request."

user_message = Prompt("""
Take a look at the following pull request:

<pull_request>
{{ title | wrap_xml('title') }}
{{ body | wrap_xml('body') }}
{{ patch | wrap_xml('patch') }}
</pull_request>

Suggest areas of the code that can be improved, such as adding unit tests for newly created utilities or features. Do not suggest refactors. If there is nothing relevant to suggest, just say "No suggestions."

Respond with this format:

<suggestions>
<suggestion>
[ Example: Add unit tests for function `foo` that tests for the case where `bar` is `baz`. ]
</suggestion>
[... other suggestions if applicable]

If there are no suggestions, just say "No suggestions."
</suggestions>
""")

def generate_auto_suggestions(pull_request: PullRequest) -> list[str]:
    # _pending_runs, _successful_runs, error_runs = get_ci_failures(pull_request)
    # logs_list = get_ci_failure_logs(installation_id, error_runs, pull_request)

    thread = Thread.from_system_message_string(system_message)
    patch = "\n".join([
        f"#### `{pull_request_file.filename}`\n```diff\n{remove_whitespace_changes(pull_request_file.patch)}\n```\n"
        for pull_request_file in pull_request.get_files() if pull_request_file.patch is not None
    ])
    response = continuous_llm_calls(
        username="suggestions", # TODO: use actual username
        thread=thread,
        content=user_message.render(
            title=pull_request.title,
            body=pull_request.body,
            patch=patch,
        ),
        stop_sequences=["</suggestion>"],
        model=LATEST_CLAUDE_MODEL,
    )
    results = extract_all_xml_tags(response, "suggestion")
    return results

# Technically we check with @sweep but @sweepai auto-corrects better
comment_template = Prompt("""
Hey @{{username}}, here {% if suggestions|length == 1 %}is an example{% else %}are examples{% endif %} of how you can ask me to improve this pull request:

{% for suggestion in suggestions %}
<blockquote>
@sweep {{ suggestion.replace('\n', '<br/>') }}
</blockquote>
{% endfor %}

""")

def provide_auto_suggestions(pull_request: PullRequest, tracking_id: str | None = None) -> list[str]:
    username = pull_request.user.login
    
    if not wildcard_match(username, FIX_CI_ALLOWLIST):
        return []
    
    message_marker = create_message_marker("auto_suggestions")

    with log_to_file(f"auto-suggestions-{tracking_id}"):
        logger.info(f"Generating auto-suggestions for {pull_request.number}")
        sweeps_github_comment = find_sweeps_github_comment(pull_request, message_marker)
        if sweeps_github_comment and (datetime.now(pytz.utc) - sweeps_github_comment.updated_at).seconds < 60:
            # If it's been recently edited, debounce. We have this due to check_suite events hammering all at once.
            return []

        suggestions = generate_auto_suggestions(pull_request)

        if not suggestions:
            return []

        # Adding CI to list of suggestions
        org_name, repo_name = pull_request.base.repo.full_name.split("/")
        installation_id = github_integration.get_repo_installation(org_name, repo_name).id

        pending_runs, _successful_runs, error_runs = get_ci_failures(pull_request)
        logs_list = get_ci_failure_logs(installation_id, error_runs, pull_request)

        if logs_list:
            suggestions = [f"Fix the CI errors."] + suggestions

        # Sending to user
        sweeps_github_comment = create_or_edit_sweeps_github_comment(pull_request, comment_template.render(
            username=username,
            suggestions=suggestions,
        ), message_marker=message_marker)
        sweeps_github_comment.create_reaction("+1")
        sweeps_github_comment.create_reaction("-1")
        return suggestions

if __name__ == "__main__":
    from sweepai.core.github_links import parse_url
    from sweepai.core.github_utils import get_github_repo
    pull_request_url = "https://github.com/sweepai/sweep-internal/pull/1315/files"
    parsed_url = parse_url(pull_request_url)

    repo = get_github_repo(f"{parsed_url.org_name}/{parsed_url.repo_name}")
    pull_request = repo.get_pull(parsed_url.pull_request_number)
    print(provide_auto_suggestions(pull_request))



from datetime import datetime
import os
import subprocess
import time
from urllib.parse import urlparse

from loguru import logger
import typer
os.environ["CLI"] = "true"

from typing import Literal
from pydantic import BaseModel
import requests

from sweepai.config.server import DEV
from sweepai.dataclasses.code_suggestions import CodeSuggestion

subprocess_args: dict = {"capture_output": True, "text": True, "check": True}

def compare_changes(changes1: list[CodeSuggestion], changes2: list[CodeSuggestion]):
    return all(change1.file_path == change2.file_path and change1.original_code == change2.original_code and change1.new_code == change2.new_code for change1, change2 in zip(changes1, changes2))

class RemoteChanges(BaseModel):
    applied_changes: list[CodeSuggestion]
    branch: str
    updated_at: datetime

    @classmethod
    def from_raw_data(cls, data: dict):
        return cls(applied_changes=[CodeSuggestion.from_camel_case(change) for change in data["applied_changes"]], branch=data["branch"], updated_at=datetime.fromisoformat(data["updated_at"]))

class SyncManager(BaseModel):
    message_id: str
    github_token: str
    latest_changes: list[CodeSuggestion] | None = None
    branch: str = ""
    updated_at: datetime | None = None
    base_url: str = "http://localhost:8080" if DEV else "https://backend.app.sweep.dev"

    @classmethod
    def from_url(cls, url: str):
        parsed_url = urlparse(url)
        _c, message_id = parsed_url.path.strip('/').split('/')
        instance = cls(message_id=message_id, github_token=os.environ["GITHUB_PAT"])
        instance.branch = subprocess.run(["git", "rev-parse", "--abbrev-ref", "HEAD"], **subprocess_args).stdout.strip("\n")
        return instance

    def _request(self, method: Literal["GET", "POST"], path: str, **kwargs):
        response = requests.request(method, f"{self.base_url}/backend/{path}", headers={"Authorization": f"Bearer {self.github_token}"}, **kwargs)
        response.raise_for_status()
        return response.json()

    def _fetch_remote_changes(self):
        raw_data = self._request("GET", f"messages/{self.message_id}/changes")
        return RemoteChanges.from_raw_data(raw_data)

    def _apply_changes(self, changes: list[CodeSuggestion]):
        num_changes = 0
        for change in changes:
            with open(change.file_path, "r") as f:
                original_code = f.read()
                # if it's original code, apply the change
                # if it's already the new code, do nothing
                if original_code not in (change.original_code, change.new_code):
                    # otherwise, ask the user
                    if not typer.confirm(f"File {change.file_path} has changed since last sync. Applying the change will overwrite it. Continue?"):
                        continue
            with open(change.file_path, "w") as f:
                f.write(change.new_code)
            num_changes += 1
        return num_changes

    def pull_changes(self):
        remote_changes = self._fetch_remote_changes()
        if self.branch != remote_changes.branch:
            raise Exception(f"Branch mismatch, skipping sync. Expected {self.branch}, got {remote_changes.branch}")
        if self.updated_at is None or remote_changes.updated_at > self.updated_at:
            logger.info(f"Pulling changes from server, {len(remote_changes.applied_changes)} changes applied (time: {remote_changes.updated_at}).")
            n = self._apply_changes(remote_changes.applied_changes)
            self.updated_at = remote_changes.updated_at
            return n
        else:
            return 0
    
    def get_local_changes(self):
        # get list of changed files before and after the commit
        files_changed = subprocess.run(["git", "diff", "--name-only", "HEAD"], **subprocess_args).stdout.strip("\n").split("\n")
        changes: list[CodeSuggestion] = []
        for file_changed in files_changed:
            # get original code before uncommitted changes
            original_code = subprocess.run(["git", "show", f"HEAD:{file_changed}"], **subprocess_args).stdout or ""
            if not os.path.exists(file_changed):
                logger.error(f"File does not exist: {file_changed}")
                continue
            with open(file_changed, "r") as f:
                new_code = f.read()
            changes.append(CodeSuggestion(file_path=file_changed, original_code=original_code, new_code=new_code))
        return changes
    
    def write_changes(self):
        changes = self.get_local_changes()
        if self.latest_changes is not None and compare_changes(self.latest_changes, changes):
            # Don't update if no changes
            return
        logger.info(f"Writing {len(changes)} changes to server.")
        self._request("POST", f"messages/{self.message_id}/changes", json=[change.to_camel_case() for change in changes])
        self.latest_changes = changes

    def disconnect(self):
        self._request("GET", f"messages/{self.message_id}/changes?disconnected=true")

    def sync_loop(self):
        # check that local branch is clean
        logger.info("Polling for changes, press Ctrl+C to stop.")
        self.pull_changes()
        self.latest_changes = self.get_local_changes()
        try:
            while True:
                self.pull_changes()
                self.write_changes()
                time.sleep(1)
        except KeyboardInterrupt:
            logger.info("Stopping sync loop.")
            self.disconnect()
            raise

def main(url: str):
    sync_manager = SyncManager.from_url(url)
    sync_manager.sync_loop()

if __name__ == "__main__":
    typer.run(main)